"""Utility functions for object detection evaluation."""

from dataclasses import dataclass
from typing import Union

import numpy as np

# NOTE: we need to add shapely to the requirements once this is enabled
from shapely import geometry

from zod.anno.object import ObjectAnnotation
from zod.data_classes.box import Box2D, Box3D
from zod.eval.detection._nuscenes_eval.detection.data_classes import DetectionBox


@dataclass
class PredictedObject:
    """Class to store dynamic object prediction information."""

    # TODO: maybe move to zod.eval?

    name: str
    confidence: float
    box3d: Box3D

    def __eq__(self, __o: Union[ObjectAnnotation, "PredictedObject"]) -> bool:
        return self.box3d == __o.box3d


def convert_to_detection_box(frame_id: str, obj: ObjectAnnotation) -> DetectionBox:
    """Convert a DynamicObject to a DetectionBox."""
    det_box = DetectionBox(
        sample_token=frame_id,
        translation=tuple(obj.box3d.center),
        size=tuple(obj.box3d.size),
        rotation=tuple(obj.box3d.orientation.elements),
        num_pts=10,
        detection_name=obj.name,
        detection_score=-1.0,  # ground truth boxes have no score
    )

    return det_box


def center_distance(box1: Box3D, box2: Box3D) -> float:
    """Calculate the center distance between two boxes."""
    return np.linalg.norm(box1.center - box2.center)


def center_distance_bev(box1: Box3D, box2: Box3D) -> float:
    """Calculate the center distance between two boxes."""
    assert box1.frame == box2.frame, "Boxes must be in the same frame."
    return np.linalg.norm(box1.center[:2] - box2.center[:2])


def iou3D(box1: Box3D, box2: Box3D) -> float:
    """Calculate the iou between two boxes."""
    assert box1.frame == box2.frame, "Boxes must be in the same frame."

    raise NotImplementedError


def iou2D(box1: Box2D, box2: Box2D) -> float:
    """Calculate the iou between two boxes."""

    # calculate the intersection
    xmin = max(box1.xyxy[0], box2.xyxy[0])
    ymin = max(box1.xyxy[1], box2.xyxy[1])
    xmax = min(box1.xyxy[2], box2.xyxy[2])
    ymax = min(box1.xyxy[3], box2.xyxy[3])
    intersection = max(0, xmax - xmin) * max(0, ymax - ymin)

    # calculate the union
    union = box1.area + box2.area - intersection

    return intersection / union if union > 0.0 else 0.0


def giou2D(box1: Box2D, box2: Box2D) -> float:
    """Calculate the giou between two boxes."""

    # calculate the intersection
    xmin = max(box1.xyxy[0], box2.xyxy[0])
    ymin = max(box1.xyxy[1], box2.xyxy[1])
    xmax = min(box1.xyxy[2], box2.xyxy[2])
    ymax = min(box1.xyxy[3], box2.xyxy[3])
    intersection = max(0, xmax - xmin) * max(0, ymax - ymin)

    # calculate the union
    union = box1.area + box2.area - intersection

    # calculate the convex hull
    xmin = min(box1.xyxy[0], box2.xyxy[0])
    ymin = min(box1.xyxy[1], box2.xyxy[1])
    xmax = max(box1.xyxy[2], box2.xyxy[2])
    ymax = max(box1.xyxy[3], box2.xyxy[3])
    convex_hull = (xmax - xmin) * (ymax - ymin)

    return intersection / union - (convex_hull - union) / convex_hull if union > 0.0 else 0.0


def iod2D(box1: Box2D, detection: Box2D) -> float:
    """Calculate the intersection over detection area between two boxes."""

    # calculate the intersection
    xmin = max(box1.xyxy[0], detection.xyxy[0])
    ymin = max(box1.xyxy[1], detection.xyxy[1])
    xmax = min(box1.xyxy[2], detection.xyxy[2])
    ymax = min(box1.xyxy[3], detection.xyxy[3])
    intersection = max(0, xmax - xmin) * max(0, ymax - ymin)

    # calculate the union
    return intersection / detection.area if detection.area > 0.0 else 0.0


def polygon_iod2D(polygon1: geometry.Polygon, detection: geometry.Polygon) -> float:
    """Calculate the intersection over detection area between two polygons."""
    intersection = polygon1.intersection(detection).area
    return intersection / detection.area if detection.area > 0.0 else 0.0


def polygon_iou2D(polygon1: geometry.Polygon, polygon2: geometry.Polygon) -> float:
    """Calculate the iou between two polygons."""
    intersection = polygon1.intersection(polygon2).area
    union = polygon1.union(polygon2).area
    return intersection / union if union > 0.0 else 0.0


def polygon_giou2D(polygon1: geometry.Polygon, polygon2: geometry.Polygon) -> float:
    """Calculate the giou between two polygons."""
    intersection = polygon1.intersection(polygon2).area
    union = polygon1.union(polygon2).area
    convex_hull = polygon1.convex_hull.union(polygon2.convex_hull).area
    return intersection / union - (convex_hull - union) / convex_hull if union > 0.0 else 0.0



from __future__ import annotations

from dataclasses import dataclass
from typing import Any, Dict, Optional

import numpy as np
from pyquaternion import Quaternion

from zod.constants import Camera, Lidar
from zod.data_classes.box import Box2D, Box3D

OBJECT_CLASSES_DYNAMIC = (
    "Vehicle",
    "VulnerableVehicle",
    "Pedestrian",
    "Animal",
)
OBJECT_CLASSES_STATIC = (
    "PoleObject",
    "TrafficBeacon",
    "TrafficSign",
    "TrafficSignal",
    "TrafficGuide",
    "DynamicBarrier",
)
OBJECT_CLASSES = (
    *OBJECT_CLASSES_DYNAMIC,
    *OBJECT_CLASSES_STATIC,
    "Unclear",
)

CLASSES_WITH_SUBCLASSES = ("Vehicle", "VulnerableVehicle", "TrafficSign", "TrafficSignal")
VEHICLE_SUBCLASSES = (
    "Car",
    "Van",
    "Truck",
    "Bus",
    "Trailer",
    "TramTrain",
    "HeavyEquip",
    "Emergency",
    "Other",
)
VULNERABLE_VEHICLE_SUBCLASSES = (
    "Bicycle",
    "Motorcycle",
    "Stroller",
    "Wheelchair",
    "PersonalTransporter",
    "NoRider",
    "Other",
)
OBJECT_SUBCLASSES = (
    *[cls_ for cls_ in OBJECT_CLASSES if cls_ not in CLASSES_WITH_SUBCLASSES],
    *[f"Vehicle_{type_}" for type_ in VEHICLE_SUBCLASSES],
    *[f"VulnerableVehicle_{type_}" for type_ in VULNERABLE_VEHICLE_SUBCLASSES],
    "TrafficSign_Front",
    "TrafficSign_Back",
    "TrafficSignal_Front",
    "TrafficSignal_Back",
)


@dataclass
class ObjectAnnotation:
    """Class to store dynamic object information."""

    # These are always available
    box2d: Box2D
    unclear: bool
    name: str
    uuid: str

    # These are not set if the object is unclear
    box3d: Optional[Box3D]  # This can be None even if the object is not unclear
    object_type: Optional[str]
    occlusion_level: Optional[str]
    artificial: Optional[str]
    with_rider: Optional[bool]
    emergency: Optional[bool]
    traffic_content_visible: Optional[bool]

    @property
    def superclass(self):
        """Get the super-class of the object."""
        return self.name

    @property
    def subclass(self):
        """Get the sub-class of the object."""
        if self.unclear:
            return f"Unclear"
        superclass = self.superclass
        if superclass not in CLASSES_WITH_SUBCLASSES:
            return superclass
        if self.artificial is True:
            return f"{superclass}_Artificial"
        if self.traffic_content_visible is not None:
            return f"{superclass}_{'Front' if self.traffic_content_visible else 'Back'}"
        if self.with_rider is False:
            return f"{superclass}_NoRider"
        if self.emergency is True and superclass == "Vehicle":
            return f"{superclass}_Emergency"  # other superclasses have too few emergency objs
        if self.object_type is not None:
            return f"{superclass}_{self.object_type}"
        # the above should have been exhaustive
        raise ValueError("Object subclass could not be determined")

    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> ObjectAnnotation:
        """Create an ObjectAnnotation from a dictionary."""
        properties: Dict[str, Any] = data["properties"]

        box2d = Box2D.from_points(points=data["geometry"]["coordinates"], frame=Camera.FRONT)
        box3d = None
        if "location_3d" in properties:
            box3d = Box3D(
                center=np.array(properties["location_3d"]["coordinates"]),
                size=np.array(
                    [
                        properties["size_3d_length"],
                        properties["size_3d_width"],
                        properties["size_3d_height"],
                    ]
                ),
                orientation=Quaternion(
                    properties["orientation_3d_qw"],
                    properties["orientation_3d_qx"],
                    properties["orientation_3d_qy"],
                    properties["orientation_3d_qz"],
                ),
                frame=Lidar.VELODYNE,
            )

        def _parse_bool_prop(key: str) -> Optional[bool]:
            return None if key not in properties else properties[key] == "True"

        superclass, obj_type = properties["class"], properties.get("type")
        unclear = "Inconclusive" in (superclass, obj_type) or properties["unclear"]
        if superclass == "Inconclusive":
            superclass = "Unclear"
        return cls(
            uuid=properties["annotation_uuid"],
            box2d=box2d,
            box3d=box3d,
            unclear=unclear,
            name=superclass,
            object_type=obj_type,
            occlusion_level=properties.get("occlusion_ratio", None),
            artificial=properties.get("artificial", None),
            with_rider=_parse_bool_prop("with_rider"),
            emergency=properties.get("emergency", None),
            traffic_content_visible=_parse_bool_prop("traffic_content_visible"),
        )


# Compatibility with old naming
AnnotatedObject = ObjectAnnotation



from __future__ import annotations

import json
from dataclasses import dataclass
from datetime import datetime, timezone
from typing import Tuple, Union

import h5py
import numpy as np
import quaternion
from scipy.interpolate import interp1d

OXTS_TIMESTAMP_OFFSET = datetime(1980, 1, 6, tzinfo=timezone.utc).timestamp()


@dataclass
class EgoMotion:
    poses: np.ndarray  # Nx4x4
    velocities: np.ndarray  # Nx3
    accelerations: np.ndarray  # Nx3
    angular_rates: np.ndarray  # Nx3
    timestamps: np.ndarray  # N - epoch time in seconds (float64)
    origin_lat_lon: Tuple[float, float]  # (lat, lon) of the origin (which pose are relative to)

    def __len__(self) -> int:
        return len(self.timestamps)

    def get_poses(self, target_ts: Union[np.ndarray, float]) -> np.ndarray:
        """Interpolate neighbouring poses to find pose for each target timestamp.
        Args:
            target_ts: [N] timestamps for which to find a pose by interpolation.
        Return:
            [N, 4, 4] array of interpolated poses for each timestamp.
        """
        selfmin, selfmax = np.min(self.timestamps), np.max(self.timestamps)
        assert (selfmin <= np.min(target_ts)) and (
            selfmax >= np.max(target_ts)
        ), f"targets not between pose timestamps, must be [{selfmin}, {selfmax}]"

        if np.isin(target_ts, self.timestamps).all():
            return self.poses[self.timestamps.searchsorted(target_ts)]

        closest_idxs = self.timestamps.searchsorted(target_ts, side="right", sorter=self.timestamps.argsort())

        # if the target timestamp is exactly the same as the largest timestamp
        # then the searchsorted will return the length of the array, which is
        # out of bounds. The assert above ensures that the target timestamp is
        # less or equal than the largest timestamp, so we can just clip the index
        # to the last element.
        closest_idxs = np.clip(closest_idxs, 1, len(self.timestamps) - 1)

        time_diffs = target_ts - self.timestamps[closest_idxs - 1]
        total_times = self.timestamps[closest_idxs] - self.timestamps[closest_idxs - 1]
        fractions = time_diffs / total_times
        return interpolate_transforms(self.poses[closest_idxs - 1], self.poses[closest_idxs], fractions)

    def interpolate(self, timestamps: np.ndarray) -> EgoMotion:
        """Interpolate ego motion to find ego motion for each target timestamp.

        Args:
            timestamps: [N] timestamps for which to find ego motion by interpolation.
        Return:
            EgoMotion object with interpolated ego motion for each timestamp.
        """
        poses = self.get_poses(timestamps)
        velocities = interpolate_vectors(self.velocities, self.timestamps, timestamps)
        accelerations = interpolate_vectors(self.accelerations, self.timestamps, timestamps)
        angular_rates = interpolate_vectors(self.angular_rates, self.timestamps, timestamps)

        return EgoMotion(
            poses=poses,
            velocities=velocities,
            accelerations=accelerations,
            angular_rates=angular_rates,
            timestamps=timestamps,
            origin_lat_lon=self.origin_lat_lon,
        )

    @classmethod
    def from_oxts_path(cls, file_path: str) -> EgoMotion:
        """Load ego motion from a sequence or frame oxts file."""
        with h5py.File(file_path, "r") as file:
            return cls(
                poses=file["poses"][()],
                accelerations=np.stack([file["accelerationX"], file["accelerationY"], file["accelerationZ"]], axis=1),
                velocities=np.stack([file["velForward"][()], file["velLateral"][()], -file["velDown"][()]], axis=1),
                angular_rates=np.stack([file["angularRateX"], file["angularRateY"], file["angularRateZ"]], axis=1),
                timestamps=OXTS_TIMESTAMP_OFFSET + file["timestamp"][()] + file["leapSeconds"][()],
                origin_lat_lon=(file["posLat"][0], file["posLon"][0]),
            )

    @classmethod
    def from_json_path(cls, json_path: str) -> EgoMotion:
        """Load ego motion from a json file."""
        with open(json_path, "r") as file:
            data = json.load(file)
        return cls(
            poses=np.array(data["poses"]),
            velocities=np.array(data["velocities"]),
            accelerations=np.array(data["accelerations"]),
            angular_rates=np.array(data["angular_rates"]),
            timestamps=np.array(data["timestamps"]),
            origin_lat_lon=tuple(data["origin_lat_lon"]),
        )

    def to_json(self) -> dict:
        """Save ego motion to a json file."""
        return {
            "poses": self.poses.tolist(),
            "velocities": self.velocities.tolist(),
            "accelerations": self.accelerations.tolist(),
            "angular_rates": self.angular_rates.tolist(),
            "timestamps": self.timestamps.tolist(),
            "origin_lat_lon": self.origin_lat_lon,
        }


def interpolate_transforms(transform_1: np.ndarray, transform_2: np.ndarray, fractions: np.ndarray):
    """Interpolate between two transforms.

    Args:
        transform_1: First transform (Mx4x4).
        transform_2: Second transform (Mx4x4).
        fraction: Fraction of interpolation (M).

    Returns:
        Interpolated transform(s) (Mx4x4).
    """
    R_a, t_a = transform_1[..., :3, :3], transform_1[..., :3, 3]
    R_b, t_b = transform_2[..., :3, :3], transform_2[..., :3, 3]
    # Alternative would be to convert to euler angles and interpolate, but this is more accurate.
    R_interpolated = quaternion.as_rotation_matrix(
        np.slerp_vectorized(
            quaternion.from_rotation_matrix(R_a),
            quaternion.from_rotation_matrix(R_b),
            fractions,
        )
    )
    t_interp = t_a + fractions[..., None] * (t_b - t_a)
    transform = np.concatenate(
        [
            np.concatenate([R_interpolated, t_interp[..., None]], axis=-1),
            np.broadcast_to(np.array([0, 0, 0, 1]), R_interpolated.shape[:-2] + (1, 4)),
        ],
        axis=-2,
    )
    return transform


def interpolate_vectors(values: np.ndarray, source_timestamps: np.ndarray, target_timestamps: np.ndarray) -> np.ndarray:
    """Interpolate vectors to find vector for each (sorted) target timestamp.

    Args:
        values: [N, X] array of values.
        source_timestamps: [N] array of timestamps.
        target_timestamps: [M] array of timestamps.
    Returns:
        [M, X] array of interpolated values.
    """
    source_min_ts, source_max_ts = np.min(source_timestamps), np.max(source_timestamps)
    target_min_ts, target_max_ts = np.min(target_timestamps), np.max(target_timestamps)
    assert source_min_ts <= target_min_ts, "Target timestamps must be after source timestamps"
    assert source_max_ts >= target_max_ts, "Target timestamps must be before source timestamps"
    # Interpolate values to find values for each target timestamp without scipy
    return interp1d(source_timestamps, values, axis=0)(target_timestamps)



# raise NotImplementedError("This evaluation framework is not yet implemented")

from itertools import repeat
from typing import Callable, Dict, Iterator, List, Tuple

import numpy as np
from tqdm.contrib.concurrent import process_map

from zod.anno.object import ObjectAnnotation
from zod.data_classes.calibration import Calibration
from zod.eval.detection._experimental.matching import MatchedFrame
from zod.eval.detection._experimental.utils import PredictedObject
from zod.eval.detection.constants import EVALUATION_CLASSES

EvalFrame = Tuple[List[ObjectAnnotation], List[PredictedObject], Calibration]


class Evalset:
    def __init__(self):
        self._frames: Dict[str, EvalFrame] = {}

    def add_frame(self, frame_id: str, frame: EvalFrame):
        self._frames[frame_id] = frame

    def __len__(self):
        return len(self._frames)

    def __repr__(self) -> str:
        return f"Evalset with {len(self)} frames."

    def __iter__(self) -> Iterator[Tuple[str, EvalFrame]]:
        return iter(self._frames.items())

    def __getitem__(self, frame_id: str) -> EvalFrame:
        return self._frames[frame_id]

    @property
    def frames(self) -> Dict[str, EvalFrame]:
        """Return all frames."""
        return self._frames


def _evaluate_frame(
    frame_id: str,
    frame: EvalFrame,
    cls: str,
    dist_fcn: Callable,
    dist_th: float,
) -> Dict:  # Dict[str, MatchedFrame]:
    from zod.eval.detection.matching import MatchedFrame, match_one_frame

    gts, preds, calibration = frame
    # filter out all predictions that are not of the correct class
    preds = [pred for pred in preds if pred.name == cls]
    # filter the ground truth boxes based on cls
    gts = [gt for gt in gts if gt.name == cls]

    matched_frame = match_one_frame(
        ground_truth=gts,
        predictions=preds,
        calibration=calibration,
        dist_fcn=dist_fcn,
        dist_threshold=dist_th,
        eval_classes=EVALUATION_CLASSES,
    )

    return {frame_id: matched_frame}


def evaluate(
    evalset: Evalset,
    dist_fcn: Callable,
    dist_th: float,
):
    # evaluate over all classes
    for cls in EVALUATION_CLASSES:
        # process all frames in parallel
        results: List[Dict[str, MatchedFrame]] = process_map(
            _evaluate_frame,
            evalset.frames.keys(),
            evalset.frames.values(),
            repeat(cls),
            repeat(dist_fcn),
            repeat(dist_th),
            desc=f"Accumulating metrics for class: {cls}",
            chunksize=1 if len(evalset.frames) < 100 else 100,
        )

        # join the list of dicts into a single dict
        results: Dict[str, MatchedFrame] = {k: v for d in results for k, v in d.items()}

        # accumulate the metrics across all frames
        tps = []
        fps = []
        confs = []
        n_ground_truths = 0

        for _, matched_frame in results.items():
            n_ground_truths += len(matched_frame.matches) + len(matched_frame.false_negatives)
            for _, pred in matched_frame.matches:
                tps.append(1)
                fps.append(0)
                confs.append(pred.confidence)
            for pred in matched_frame.false_positives:
                tps.append(0)
                fps.append(1)
                confs.append(pred.confidence)

        # sort the accumulated list based on confidence
        confs, tps, fps = zip(*sorted(zip(confs, tps, fps), reverse=True))
        # accumulate the tps and fps
        tps = np.cumsum(tps).astype(np.float)
        fps = np.cumsum(fps).astype(np.float)
        # compute the precision and recall
        precision = tps / (tps + fps)
        recall = tps / n_ground_truths

        raise NotImplementedError("AP computation is not yet implemented")
        # compute the ap
        recall_sampling_points = np.linspace(0, 1, PRECISION_RECALL_SAMPLING_POINTS)
        precision = np.interp(recall_sampling_points, recall, precision, right=0)
        # confidences = np.interp(recall_sampling_points, recall, confs, right=0)
        recall = recall_sampling_points

        # store in some sort of data structure, both the raw data but also the computed metrics
        # todo (william), store the results here
        ap = np.mean(precision)
        print(f"AP for class {cls} is {ap}")

    # compute the final metrics, average over classes etc
    # todo (william), compute the final metrics here



"""ZOD Frames module."""

import json
import os.path as osp
from abc import ABC, abstractmethod
from functools import partial
from itertools import chain
from typing import Any, Dict, Optional, Set, Tuple, Union

from tqdm.contrib.concurrent import process_map

from zod.constants import FULL, TRAIN, VAL, VERSIONS, AnnotationProject, Split, Version
from zod.data_classes.info import Information
from zod.utils.utils import zfill_id


def _create_frame(frame: dict, dataset_root: str) -> Information:
    info = Information.from_dict(frame)
    info.convert_paths_to_absolute(dataset_root)
    return info


class ZodDataset(ABC):
    """Base class for all ZOD sub-datasets (frames, sequences, drives)."""

    def __init__(self, dataset_root: str, version: Version, mp: bool = True):
        self._dataset_root = dataset_root
        self._version = version
        self._mp = mp
        assert version in VERSIONS, f"Unknown version: {version}, must be one of: {VERSIONS}"
        self._infos, self._train_ids, self._val_ids = self._load_infos()

    def __len__(self) -> int:
        return len(self._infos)

    def __iter__(self):
        for frame_id in self._infos:
            if frame_id not in self._train_ids and frame_id not in self._val_ids:
                continue
            yield self.__getitem__(frame_id)

    @abstractmethod
    def __getitem__(self, frame_id: Union[int, str, slice]) -> Any:
        """Get item by id, which is a 6-digit zero-padded number. Ex: '000001'."""
        if isinstance(frame_id, slice):
            return [self.__getitem__(i) for i in range(*frame_id.indices(len(self)))]
        frame_id = zfill_id(frame_id)
        return self._infos[frame_id]

    @property
    @abstractmethod
    def trainval_files(self) -> Dict[str, str]:
        """Mapping of version to trainval file."""
        raise NotImplementedError

    def get_split(self, split: Split, project: Optional[AnnotationProject] = None) -> Set[str]:
        """Get split by name (e.g. train / val)."""
        if split == TRAIN:
            ids = self._train_ids
        elif split == VAL:
            ids = self._val_ids
        else:
            raise ValueError(f"Unknown split: {split}, should be {TRAIN} or {VAL}")

        if project is None:
            return ids
        else:
            return {id_ for id_ in ids if project in self._infos[id_].annotations}

    def get_all_infos(self) -> Dict[str, Information]:
        """Get all infos (including blacklisted ones)."""
        return self._infos

    def get_all_ids(self) -> Set[str]:
        """Get all frame ids (excluding blackisted ones)."""
        return self._train_ids.union(self._val_ids)

    def _load_infos(self) -> Tuple[Dict[str, Information], Set[str], Set[str]]:
        """Load frames for the given version."""
        trainval_path = osp.join(self._dataset_root, self.trainval_files[self._version])
        if not osp.exists(trainval_path):
            msg = f"Could not find trainval file: {trainval_path}.\n"
            if osp.exists(trainval_path.replace("-", "_")):
                msg += (
                    "However, found old, incompatible trainval files. Please either downgrade zod "
                    "to < 0.2 or download new files with `zod download --no-images --no-lidar`"
                )
            else:
                cls_name = f"{self.__class__.__name__}-{self._version}"
                msg += f"Make sure you have downloaded {cls_name} to {self._dataset_root}"
            raise FileNotFoundError(msg)
        with open(trainval_path, "r") as f:
            all_ids = json.load(f)

        func = partial(_create_frame, dataset_root=self._dataset_root)
        if self._mp and self._version == FULL:
            infos = process_map(
                func,
                chain.from_iterable(all_ids.values()),
                desc="Loading infos",
                chunksize=50 if self._version == FULL else 1,
            )
            infos = {frame.id: frame for frame in infos}
        else:
            infos = {frame.id: frame for frame in map(func, chain.from_iterable(all_ids.values()))}

        train_ids = {f["id"] for f in all_ids[TRAIN]}
        val_ids = {f["id"] for f in all_ids[VAL]}
        return infos, train_ids, val_ids



"""ZOD dataclasses."""

from __future__ import annotations

from dataclasses import dataclass
from datetime import datetime
from typing import Any, List, Union

import numpy as np
from PIL import Image

from zod.anno.parser import ANNOTATION_PARSERS
from zod.constants import AnnotationProject
from zod.utils.utils import parse_datetime_from_filename

from ._serializable import JSONSerializable
from .geometry import Pose

# we want to remove the points that get returns from the ego vehicle.
# we do this by removing all points that are within the box defined by
# EGO_RETURNS_BOX. This box is defined in the LiDAR frame.
EGO_RETURNS_BOX = np.array([[-1.5, -3.0, -1.5], [1.5, 3.0, 0.5]])


@dataclass
class LidarData:
    """A class describing the lidar data."""

    points: np.ndarray  # (N, 3) float32
    timestamps: np.ndarray  # (N,) float64 epoch timestamp pointwise
    intensity: np.ndarray  # (N,) uint8
    diode_idx: np.ndarray  # (N,) uint8
    core_timestamp: float  # core epoch timestamp

    def copy(self) -> LidarData:
        """Return a copy of the lidar data."""
        return LidarData(
            points=self.points.copy(),
            timestamps=self.timestamps.copy(),
            intensity=self.intensity.copy(),
            diode_idx=self.diode_idx.copy(),
            core_timestamp=self.core_timestamp,
        )

    @classmethod
    def empty(cls) -> LidarData:
        """Create an empty lidar data object."""
        return cls(
            points=np.empty((0, 3), dtype=np.float32),
            timestamps=np.empty(0, dtype=np.float64),
            intensity=np.empty(0, dtype=np.uint8),
            diode_idx=np.empty(0, dtype=np.uint8),
            core_timestamp=0.0,
        )

    @classmethod
    def from_npy(cls, path: str, remove_ego_lidar_returns: bool = True) -> LidarData:
        """Load lidar data from a .npy file."""
        data = np.load(path)
        core_timestamp = parse_datetime_from_filename(path).timestamp()
        if remove_ego_lidar_returns:
            ego_returns_mask = np.ones(len(data), dtype=np.bool_)
            for idim, dim in enumerate(("x", "y", "z")):
                ego_returns_mask &= np.logical_and(
                    data[dim] > EGO_RETURNS_BOX[0, idim], data[dim] < EGO_RETURNS_BOX[1, idim]
                )
            data = data[~ego_returns_mask]
        return cls(
            points=np.vstack((data["x"], data["y"], data["z"])).T,
            timestamps=core_timestamp + data["timestamp"] / 1e6,
            intensity=data["intensity"],
            diode_idx=data["diode_index"],
            core_timestamp=core_timestamp,
        )

    def to_npy(self, path: str) -> None:
        """Save lidar data to a .npy file in the same format as is used for loading."""
        data = np.empty(
            len(self.points),
            dtype=[
                ("x", np.float32),
                ("y", np.float32),
                ("z", np.float32),
                ("timestamp", np.int64),
                ("intensity", np.uint8),
                ("diode_index", np.uint8),
            ],
        )
        data["x"] = self.points[:, 0]
        data["y"] = self.points[:, 1]
        data["z"] = self.points[:, 2]
        data["timestamp"] = (self.timestamps - self.core_timestamp) * 1e6
        data["intensity"] = self.intensity
        data["diode_index"] = self.diode_idx
        np.save(path, data)

    def transform(self, pose: Union[np.ndarray, Pose]) -> None:
        """Transform the lidar data to a new pose.

        Args:
            pose: The new pose to transform the lidar data to.
        """
        if isinstance(pose, Pose):
            pose = pose.transform
        rotations = pose[..., :3, :3]
        translations = pose[..., :3, 3]
        self.points = (self.points[..., None, :] @ rotations.swapaxes(-2, -1) + translations[..., None, :]).squeeze(-2)

    def extend(self, *other: LidarData):
        """Extend this LidarData with data from another LidarData object.

        Args:
            other: The other LidarData object.
        """
        self.points = np.vstack((self.points, *(o.points for o in other)))
        self.timestamps = np.hstack((self.timestamps, *(o.timestamps for o in other)))
        self.intensity = np.hstack((self.intensity, *(o.intensity for o in other)))
        self.diode_idx = np.hstack((self.diode_idx, *(o.diode_idx for o in other)))
        # Core timestamp is the weighted average
        self.core_timestamp = (
            self.core_timestamp * len(self.timestamps) + sum(o.core_timestamp * len(o.timestamps) for o in other)
        ) / (len(self.timestamps) + sum(len(o.timestamps) for o in other))

    def __eq__(self, other: LidarData) -> Union[bool, np.bool_]:
        """Check if two LidarData objects are equal."""
        return (
            np.allclose(self.points, other.points)
            and np.allclose(self.timestamps, other.timestamps)
            and np.allclose(self.intensity, other.intensity)
            and np.allclose(self.diode_idx, other.diode_idx)
            and np.isclose(self.core_timestamp, other.core_timestamp)
        )


@dataclass
class RadarData:
    """A class describing the radar data."""

    radar_range: np.ndarray  # (N,) float32
    azimuth_angle: np.ndarray  # (N,) float32
    elevation_angle: np.ndarray  # (N,) float32
    range_rate: np.ndarray  # (N,) float32
    amplitude: np.ndarray  # (N,) float32
    validity: np.ndarray  # (N,) int8
    mode: np.ndarray  # (N,) uint8
    quality: np.ndarray  # (N,) uint8
    scan_index: np.ndarray  # (N,) uint32
    timestamp: np.ndarray  # (N,) int64

    def copy(self) -> RadarData:
        """Return a copy of the radar data."""
        return RadarData(
            radar_range=self.radar_range.copy(),
            azimuth_angle=self.azimuth_angle.copy(),
            elevation_angle=self.elevation_angle.copy(),
            range_rate=self.range_rate.copy(),
            amplitude=self.amplitude.copy(),
            validity=self.validity.copy(),
            mode=self.mode.copy(),
            quality=self.quality.copy(),
            scan_index=self.scan_index.copy(),
            timestamp=self.timestamp,
        )

    @classmethod
    def empty(cls) -> RadarData:
        """Create an empty radar data object."""
        return cls(
            radar_range=np.empty(0, dtype=np.float32),
            azimuth_angle=np.empty(0, dtype=np.float32),
            elevation_angle=np.empty(0, dtype=np.float32),
            range_rate=np.empty(0, dtype=np.float32),
            amplitude=np.empty(0, dtype=np.float32),
            validity=np.empty(0, dtype=np.int8),
            mode=np.empty(0, dtype=np.uint8),
            quality=np.empty(0, dtype=np.uint8),
            scan_index=np.empty(0, dtype=np.uint32),
            timestamp=0,
        )

    @classmethod
    def from_npy(cls, path: str) -> RadarData:
        """Load radar data from a .npy file.

        Args:
            path: Path to the file we are loading the data from."""
        data = np.load(path)
        return cls(
            radar_range=data["radar_range"],
            azimuth_angle=data["azimuth_angle"],
            elevation_angle=data["elevation_angle"],
            range_rate=data["range_rate"],
            amplitude=data["amplitude"],
            validity=data["validity"],
            mode=data["mode"],
            quality=data["quality"],
            scan_index=data["scan_index"],
            timestamp=data["timestamp"],
        )

    def to_npy(self, path: str) -> None:
        """Save radar data to a .npy file in the same format as is used for loading.

        Args:
            path: Path of the file we are saving the data in."""
        data = np.empty(
            len(self.radar_range),
            dtype=[
                ("scan_index", np.uint32),
                ("timestamp", np.int64),
                ("radar_range", np.float32),
                ("azimuth_angle", np.float32),
                ("elevation_angle", np.float32),
                ("range_rate", np.float32),
                ("amplitude", np.float32),
                ("validity", np.int8),
                ("mode", np.uint8),
                ("quality", np.uint8),
            ],
        )

        data["radar_range"] = self.radar_range
        data["azimuth_angle"] = self.azimuth_angle
        data["elevation_angle"] = self.elevation_angle
        data["range_rate"] = self.range_rate
        data["amplitude"] = self.amplitude
        data["validity"] = self.validity
        data["mode"] = self.mode
        data["quality"] = self.quality
        data["scan_index"] = self.scan_index
        if len(self.timestamp) == 1:
            data["timestamp"] = self.timestamp
        else:
            times = np.empty(len(self.radar_range), dtype=np.int64)
            for i in range(len(self.timestamp)):
                times[self.scan_index == i] = self.timestamp[i]
            data["timestamp"] = times

        np.save(path, data)

    def get_cartesian_coordinates(self) -> np.ndarray:
        """Convert radar data to cartesian coordinates with shape (N x 3)."""
        x = self.radar_range * np.cos(self.elevation_angle) * np.cos(self.azimuth_angle)
        y = self.radar_range * np.cos(self.elevation_angle) * np.sin(self.azimuth_angle)
        z = self.radar_range * np.sin(self.elevation_angle)
        return np.vstack((x, y, z)).T

    def extend(self, *other: RadarData):
        """Extend this RadarData with data from another RadarData object.

        Args:
            other: The other RadarData object.
        """
        self.radar_range = np.hstack((self.radar_range, *(o.radar_range for o in other)))
        self.azimuth_angle = np.hstack((self.azimuth_angle, *(o.azimuth_angle for o in other)))
        self.elevation_angle = np.hstack((self.elevation_angle, *(o.elevation_angle for o in other)))
        self.range_rate = np.hstack((self.range_rate, *(o.range_rate for o in other)))
        self.amplitude = np.hstack((self.amplitude, *(o.amplitude for o in other)))
        self.validity = np.hstack((self.validity, *(o.validity for o in other)))
        self.mode = np.hstack((self.mode, *(o.mode for o in other)))
        self.quality = np.hstack((self.quality, *(o.quality for o in other)))
        self.scan_index = np.hstack((self.scan_index, *(o.scan_index for o in other)))
        self.timestamp = np.vstack((self.timestamp, *(o.timestamp for o in other)))

    def __eq__(self, other: RadarData) -> Union[bool, np.bool_]:
        """Check if two RadarData objects are equal.

        Args:
            other: The other RadarData object."""
        return (
            np.allclose(self.radar_range, other.radar_range)
            and np.allclose(self.azimuth_angle, other.azimuth_angle)
            and np.allclose(self.elevation_angle, other.elevation_angle)
            and np.allclose(self.range_rate, other.range_rate)
            and np.allclose(self.amplitude, other.amplitude)
            and np.allclose(self.validity, other.validity)
            and np.allclose(self.mode, other.mode)
            and np.allclose(self.quality, other.quality)
            and np.allclose(self.scan_index, other.scan_index)
            and np.isclose(self.timestamp, other.timestamp)
        )


@dataclass
class SensorFrame(JSONSerializable):
    """Class to store sensor information."""

    filepath: str
    time: datetime

    def read(self) -> Any:
        """Read the data from the file."""
        raise NotImplementedError


@dataclass
class LidarFrame(SensorFrame):
    """Class to store information about a lidar frame."""

    is_compensated: bool  # Whether the cloud is pointwise compensated

    def read(self, remove_ego_lidar_returns: bool = True) -> LidarData:
        """Read the point cloud."""
        return LidarData.from_npy(self.filepath, remove_ego_lidar_returns=remove_ego_lidar_returns)


@dataclass
class RadarFrames(JSONSerializable):
    """Class to store information about a radar sequence file."""

    filepath: str
    time: datetime  # time of the sequence key frame

    def read(self) -> RadarData:
        """Read the radar data."""
        return RadarData.from_npy(self.filepath)


@dataclass
class CameraFrame(SensorFrame):
    """Class to store information about a camera frame."""

    height: int = 2168
    width: int = 3848

    def read(self) -> np.ndarray:
        """Read the image."""
        return np.array(Image.open(self.filepath))



import os.path as osp
from dataclasses import dataclass
from typing import Any, Dict

import h5py
import numpy as np


@dataclass
class EgoVehicleData:
    roll_rate: np.ndarray  # radians/s
    picth_rate: np.ndarray  # radians/s
    lat_vel: np.ndarray  # m/s
    lon_vel: np.ndarray  # m/s
    lat_acc: np.ndarray  # m/s^2
    lon_acc: np.ndarray  # m/s^2
    body_height: np.ndarray  # meters
    body_pitch: np.ndarray  # radians
    timestamp: np.ndarray  # epoch time, nanosec

    @classmethod
    def from_hdf5(cls, path: str) -> "EgoVehicleData":
        """Read the ego-vehicle data from the hdf5 file."""
        assert osp.exists(path), "Please download the vehicle data first."
        with h5py.File(path, "r") as h5_file:
            evd = h5_file["ego_vehicle_data"]
            return cls(
                roll_rate=evd["roll_rate_data/angle_rate/radians_per_second/value"][:],
                picth_rate=evd["pitch_rate_data/angle_rate/radians_per_second/value"][:],
                lat_vel=evd["lat_vel_data/velocity/meters_per_second/value"][:],
                lon_vel=evd["lon_vel_data/velocity/meters_per_second/value"][:],
                lat_acc=evd["lat_acc_data/acceleration/meters_per_second2/value"][:],
                lon_acc=evd["lon_acc_data/acceleration/meters_per_second2/value"][:],
                body_height=evd["body_height/body_height/meters/value"][:],
                body_pitch=evd["body_pitch/angle/radians/value"][:],
                timestamp=evd["timestamp/nanoseconds/value"][:],
            )


@dataclass
class EgoVehicleControls:
    acc_pedal: np.ndarray  # ratio, range 0-100
    brake_pedal_pressed: np.ndarray  # bool
    steering_angle: np.ndarray  # radians, counter-clockwise positive
    steering_angle_rate: np.ndarray  # radians/s
    steering_wheel_torque: np.ndarray  # Nm
    turn_indicator: np.ndarray  # 0: off, 1: left, 2: right
    timestamp: np.ndarray  # epoch time, nanosec

    @classmethod
    def from_hdf5(cls, path: str) -> "EgoVehicleControls":
        """Read the ego-vehicle control data from the hdf5 file."""
        assert osp.exists(path), "Please download the vehicle data first."
        with h5py.File(path, "r") as h5_file:
            evc = h5_file["ego_vehicle_controls"]
            return cls(
                acc_pedal=evc["acceleration_pedal/ratio/unitless/value"][:],
                brake_pedal_pressed=evc["brake_pedal_pressed/is_brake_pedal_pressed/unitless/value"][:],
                steering_angle=evc["steering_wheel_angle/angle/radians/value"][:],
                steering_angle_rate=evc["steering_wheel_angle/angle_rate/radians_per_second/value"][:],
                steering_wheel_torque=evc["steer_wheel_torque/torque/newton_meters/value"][:],
                turn_indicator=evc["turn_indicator_status/state"][:],
                timestamp=evc["timestamp/nanoseconds/value"][:],
            )


@dataclass
class Satellite:
    altitude: np.ndarray  # meters
    heading: np.ndarray  # degrees
    latpos: np.ndarray  # nano degrees
    lonpos: np.ndarray  # nano degrees
    nrof_satellites: np.ndarray  # number of satellites, integer
    speed: np.ndarray  # m/s
    timstamp: np.ndarray  # epoch time, nanosec

    @classmethod
    def from_hdf5(cls, path: str) -> "Satellite":
        """Read the sattelite data from the hdf5 file."""
        assert osp.exists(path), "Please download the vehicle data first."
        with h5py.File(path, "r") as h5_file:
            s = h5_file["satellite"]
            return cls(
                altitude=s["altitude/meters/value"][:],
                heading=s["heading/degrees/value"][:],
                latpos=s["latposn/nanodegrees/value"][:],
                lonpos=s["longposn/nanodegrees/value"][:],
                nrof_satellites=s["nrof_satellites/unitless/value"][:],
                speed=s["speed/meters_per_second/value"][:],
                timstamp=s["timestamp/nanoseconds/value"][:],
            )


@dataclass
class VehicleData:
    """Class to store the vehicle data."""

    ego_vehicle_data: EgoVehicleData
    ego_vehicle_controls: EgoVehicleControls
    satellite: Satellite

    @classmethod
    def from_hdf5(cls, path: str) -> "VehicleData":
        """Read the vehicle data from the hdf5 file."""
        assert osp.exists(path), "Please download the vehicle data first."
        return cls(
            EgoVehicleData.from_hdf5(path),
            ego_vehicle_controls=EgoVehicleControls.from_hdf5(path),
            satellite=Satellite.from_hdf5(path),
        )



"""3D visualization tool."""

from typing import Any, List, Tuple

import dash
import dash_bootstrap_components as dbc
import numpy as np
import plotly.express as px
import plotly.graph_objects as go
from dash import dcc, html

from zod import ZodFrames
from zod.constants import AnnotationProject
from zod.visualization.bev_utils import BEVSettings


def get_header(name):
    """Create an HTML header."""
    title = html.H4(name, style={"margin-top": 5})
    logo = html.Img(
        src="https://zenseact.com/wp-content/uploads/2022/10/Zenseact-logo-neg.svg",
        style={"float": "right", "height": 40, "margin-top": 5},
    )
    link = html.A(logo, href="https://zenseact.com/")

    return dbc.Row([dbc.Col(title, md=8), dbc.Col(link, md=4)], style={"height": 40})


class Viz3D:
    """3D Visualization class."""

    def __init__(self):
        """Initialize visualization figure."""
        self._settings = BEVSettings()
        self.app = dash.Dash(__name__, external_stylesheets=[dbc.themes.CYBORG])
        self.app.title = "Zenseact Open Dataset Visualization"

        layout = go.Layout(
            template="plotly_dark",
            scene=dict(aspectmode="data"),
            margin=dict(l=0, r=0, t=0, b=0),
            xaxis=dict(dtick=5, range=[-200, 200]),
            yaxis=dict(dtick=5, range=[-100, 100]),
            showlegend=True,
            legend=dict(yanchor="top", y=0.99, xanchor="left", x=0.01),
        )

        self.fig = go.Figure(layout=layout)

    def add_pointcloud(self, x, y, z, intensity):
        """Add pointcloud to 3D graph."""
        self.fig.add_traces(
            data=[
                go.Scatter3d(
                    x=x,
                    y=y,
                    z=z,
                    name="LiDAR",
                    legendgroup="LiDAR",
                    mode="markers",
                    marker=dict(
                        size=1,
                        color=intensity,  # set color to an array/list of desired values
                        colorscale="viridis",  # choose a colorscale
                        opacity=1.0,
                    ),
                    showlegend=True,
                )
            ]
        )

    def add_zod_annotations(self, objects: Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]):
        """Add 3D bounding boxed to figure."""
        class_names, positions, dimensions, rotations = objects
        n_objects = np.shape(positions)[0]

        if n_objects > 0:
            # Map classes
            classes = self._settings.encode_classes(class_names)
            for obj_idx in range(n_objects):
                self._add_box(
                    (
                        class_names[obj_idx],
                        classes[obj_idx],
                        positions[obj_idx],
                        dimensions[obj_idx],
                        rotations[obj_idx],
                    )
                )

    def _add_box(self, object_annotations: Tuple[Any], draw_edges: bool = False):
        """Construct 3D box mesh."""
        # pylint: disable-msg=too-many-locals

        # Extract object information
        class_name, class_ind, position, dimension, rotation = object_annotations

        coord1 = (-1) * dimension / 2
        coord3 = dimension / 2
        coord2 = np.vstack([coord1[0], coord3[1]])
        coord4 = np.vstack([coord3[0], coord1[1]])

        all_coords = np.vstack(
            [
                (rotation.rotate(coord1) + position),
                rotation.rotate(np.vstack([coord2, coord1[2]])) + position,
                rotation.rotate(coord3) + position,
                rotation.rotate(np.vstack([coord4, coord1[2]])) + position,
            ]
        ).T
        a = (coord1[2] + position[2]) * np.ones(4)
        b = (coord3[2] + position[2]) * np.ones(4)

        all_coords = np.hstack([np.vstack([all_coords[:2], a]), np.vstack([all_coords[:2], b])])

        colors_available = (
            px.colors.qualitative.Light24[:2]
            + px.colors.qualitative.Light24[4:8]
            + px.colors.qualitative.Light24[16:20]
            + px.colors.qualitative.Light24[-2:]
        )
        color = colors_available[int(class_ind)]

        # Boxes edges
        i = [7, 0, 0, 0, 4, 4, 6, 1, 4, 0, 3, 6]
        j = [3, 4, 1, 2, 5, 6, 5, 2, 0, 1, 6, 3]
        k = [0, 7, 2, 3, 6, 7, 1, 6, 5, 5, 7, 2]

        if not draw_edges:
            self.fig.add_traces(
                [
                    go.Mesh3d(
                        # 8 vertices of a cube
                        x=all_coords[0, :],
                        y=all_coords[1, :],
                        z=all_coords[2, :],
                        i=i,
                        j=j,
                        k=k,
                        name=class_name,
                        legendgroup=class_name,
                        legendgrouptitle=dict(text=class_name),
                        showlegend=True,
                        opacity=0.2,
                        color=color,
                        flatshading=True,
                        contour=dict(color="white", show=True, width=10),
                        alphahull=0,
                    )
                ]
            )
        else:
            triangles = np.vstack((i, j, k)).T
            x = all_coords[0, :]
            y = all_coords[1, :]
            z = all_coords[2, :]
            vertices = np.vstack((x, y, z)).T
            tri_points = vertices[triangles]
            # extract the lists of x, y, z of the triangle vertices and connect them by a line
            xe = []
            ye = []
            ze = []
            for T in tri_points:
                xe.extend([T[k % 3][0] for k in range(4)] + [None])
                ye.extend([T[k % 3][1] for k in range(4)] + [None])
                ze.extend([T[k % 3][2] for k in range(4)] + [None])

            self.fig.add_traces(
                [
                    go.Scatter3d(
                        x=xe,
                        y=ye,
                        z=ze,
                        mode="lines",
                        name="",
                        line=dict(color=color, width=14),
                    ),
                    go.Scatter3d(
                        x=xe,
                        y=ye,
                        z=ze,
                        mode="lines",
                        name="",
                        line=dict(color="black", width=2),
                    ),
                ]
            )

    def show(self):
        """Render graph."""
        self.fig.update_layout(legend_title_text="Available Objects")
        self.app.layout = dbc.Container(
            [
                get_header(self.app.title),
                html.Hr(),
                dbc.Row(dcc.Graph(id="my-graph", figure=self.fig), style={"flexGrow": "1"}),
            ],
            fluid=True,
            style={"height": "100vh", "display": "flex", "flexDirection": "column"},
        )

        self.app.run_server(debug=False)


if __name__ == "__main__":
    #
    # Zenseact Open Dataset
    #
    # NOTE! Set the dataset path
    DATASET_ROOT = ""
    VERSION = "full"  # "mini" or "full"

    # initialize ZodFrames
    zod_frames = ZodFrames(dataset_root=DATASET_ROOT, version=VERSION)

    zod_frame = zod_frames[0]

    # get the LiDAR point cloud
    pcd = zod_frame.get_lidar()[0]

    # get the object annotations
    annotations = zod_frame.get_annotation(AnnotationProject.OBJECT_DETECTION)

    viz = Viz3D()
    viz.add_pointcloud(pcd.points[:, 0], pcd.points[:, 1], pcd.points[:, 2], pcd.intensity)
    viz.add_zod_annotations(
        (
            np.array([obj.name for obj in annotations if obj.box3d]),
            np.concatenate([obj.box3d.center[None, :] for obj in annotations if obj.box3d], axis=0),
            np.concatenate([obj.box3d.size[None, :] for obj in annotations if obj.box3d], axis=0),
            np.array([obj.box3d.orientation for obj in annotations if obj.box3d]),
        )
    )

    viz.show()



"""This script is to be used to download the Zenseact Open Dataset."""

import contextlib
import os
import os.path as osp
import subprocess
import tarfile
import threading
from concurrent.futures import ThreadPoolExecutor
from dataclasses import dataclass

import requests
from tqdm import tqdm

from zod.cli.utils import SubDataset, Version

try:
    import click.exceptions
    import dropbox
    import dropbox.exceptions
    import dropbox.files
    import dropbox.sharing
    import typer
except ImportError:
    print('zod is installed without the CLI dependencies: pip install "zod[cli]"')
    exit(1)


APP_KEY = "kcvokw7c7votepo"
REFRESH_TOKEN = "z2h_5rjphJEAAAAAAAAAAUWtQlltCYlMbZ2WqMgtymELhiSuKIksxAYeArubKnZV"
CHUNK_SIZE = 8 * 1024 * 1024  # 8 MB
TIMEOUT = 8 * 60 * 60  # 8 hours

app = typer.Typer(help="Zenseact Open Dataset Downloader", no_args_is_help=True)


@dataclass
class DownloadExtractInfo:
    """Information about a file to extract."""

    url: str
    file_path: str
    extract_dir: str
    dl_dir: str
    rm: bool
    dry_run: bool
    size: int
    content_hash: str
    extract: bool
    extract_already_downloaded: bool


@dataclass
class FilterSettings:
    """Filter settings."""

    version: Version
    annotations: bool
    images: bool
    blur: bool
    dnat: bool
    lidar: bool
    oxts: bool
    infos: bool
    vehicle_data: bool
    num_scans_before: int
    num_scans_after: int

    def __str__(self):
        return "\n".join([f"    {key}: {value}" for key, value in self.__dict__.items()])


@dataclass
class DownloadSettings:
    """Download settings."""

    url: str
    output_dir: str
    rm: bool
    dry_run: bool
    extract: bool
    extract_already_downloaded: bool
    parallel: bool
    max_workers: int = 8

    def __str__(self):
        return "\n".join([f"    {key}: {value}" for key, value in self.__dict__.items()])


class ResumableDropbox(dropbox.Dropbox):
    """A patched dropbox client that allows to resume downloads."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._headers = {} if self._headers is None else self._headers
        self._lock = threading.Lock()

    def sharing_get_shared_link_file(self, url, path=None, link_password=None, start=0):
        """
        Download the shared link's file from a user's Dropbox.
        """
        # Need to lock here because the headers are shared between all requests
        with self._lock:
            self._headers["Range"] = f"bytes={start}-"
            res = super().sharing_get_shared_link_file(url, path=path, link_password=link_password)
            del self._headers["Range"]
        return res


def _print_final_msg(pbar: tqdm, basename: str):
    """Print a final message for each downloaded file, with stats and nice padding."""
    msg = "Downloaded " + basename + " " * (45 - len(basename))
    total_time = pbar.format_dict["elapsed"]
    size_in_mb = pbar.n / 1024 / 1024
    speed_in_mb = size_in_mb / total_time
    for name, val, unit in [
        ("time", total_time, "s"),
        ("size", size_in_mb, "MB"),
        ("speed", speed_in_mb, "MB/s"),
    ]:
        val = f"{val:.2f}{unit}"
        msg += f"{name}: {val}" + " " * (14 - len(val))
    tqdm.write(msg)


def _download(download_path: str, dbx: ResumableDropbox, info: DownloadExtractInfo):
    current_size = 0
    if osp.exists(download_path):
        current_size = osp.getsize(download_path)
    basename = osp.basename(info.file_path)
    pbar = tqdm(
        total=info.size,
        unit="iB",
        unit_scale=True,
        desc=f"Downloading {basename}...",
        leave=False,
        initial=current_size,
    )
    if pbar.n > info.size:
        tqdm.write(
            f"Error! File {download_path} already exists and is larger than expected. " "Please delete and try again."
        )
    if pbar.n > 0:
        # this means we are retrying or resuming a previously interrupted download
        tqdm.write(f"Resuming download of {download_path} from {current_size} bytes.")
    # Retry download if partial file exists (e.g. due to network error)
    while pbar.n < info.size:
        try:
            _, response = dbx.sharing_get_shared_link_file(url=info.url, path=info.file_path, start=pbar.n)
            with open(download_path, "ab") as f:
                with contextlib.closing(response):
                    for chunk in response.iter_content(chunk_size=CHUNK_SIZE):
                        if chunk:  # filter out keep-alive new chunks
                            size = f.write(chunk)
                            pbar.update(size)
        except requests.exceptions.ChunkedEncodingError:
            continue  # This can happen when dropbox unexpectly closes the connection
    # Final checks and printouts
    assert osp.getsize(download_path) == info.size
    _print_final_msg(pbar, basename)


def _extract(tar_path: str, output_dir: str):
    """Extract a tar file to a directory."""
    pbar = tqdm(desc=f"Extracting {osp.basename(tar_path)}...", leave=False)
    # Check if system has tar installed (pipe stderr to None to avoid printing)
    if os.system("tar --version > /dev/null 2>&1") == 0:
        with subprocess.Popen(
            ["tar", "-xvf", tar_path, "-C", output_dir],
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
        ) as tar:
            for _ in tar.stdout:
                pbar.update(1)
        # check if tar exited with error
        if tar.returncode != 0:
            tqdm.write(f"Error extracting {tar_path} with tar.")
    else:
        # Fallback to python tarfile
        with tarfile.open(name=tar_path) as tar:
            # pbar.total = len(tar.getmembers())
            for member in tar.getmembers():
                tar.extract(member=member, path=output_dir)
                pbar.update(1)
    tqdm.write(f"Extracted {pbar.n} files from {osp.basename(tar_path)}.")


def _already_downloaded(download_path: str, target_size: int):
    """Check if the file is already downloaded."""
    old_format_path = download_path.rsplit("_", 1)[0]  # /path/to/file_abc123 -> /path/to/file
    if osp.exists(download_path) and osp.getsize(download_path) == target_size:
        return True
    elif osp.exists(old_format_path) and osp.getsize(old_format_path) == target_size:
        return True
    return False


def _download_and_extract(dbx: ResumableDropbox, info: DownloadExtractInfo):
    """Download a file from a Dropbox share link to a local path."""
    should_extract = info.extract
    file_name = osp.basename(info.file_path)
    data_name = file_name.split(".")[0]
    download_path = osp.join(info.dl_dir, f"{file_name}_{info.content_hash[:8]}")
    if _already_downloaded(download_path, info.size):
        if not info.extract_already_downloaded:
            tqdm.write(f"{data_name} already exists. Skipping download and extraction.")
            should_extract = False
        else:
            tqdm.write(f"{data_name} already exists. Skipping download.")
    elif info.dry_run:
        msg = "download" if not should_extract else "download and extract"
        typer.echo(f"Would {msg} {info.file_path} to {download_path}")
        return
    else:
        try:
            _download(download_path, dbx, info)
        except Exception as e:
            print(f"Error downloading {data_name}: {e}. Please retry")
            return

    if should_extract:
        try:
            _extract(download_path, info.extract_dir)
        except Exception as e:
            print(f"Error extracting {data_name}: {e}. Please retry")
            return

    if info.rm and not info.dry_run:
        os.remove(download_path)


def _filter_entry(entry: dropbox.files.Metadata, settings: FilterSettings) -> bool:
    """Filter the entry based on the flags."""
    if settings.version == Version.MINI:
        return "mini" in entry.name
    elif "mini" in entry.name:
        return False

    if not settings.annotations and "annotations" in entry.name:
        return False
    if "images" in entry.name:
        if not settings.images:
            return False
        if not settings.blur and "blur" in entry.name:
            return False
        if not settings.dnat and "dnat" in entry.name:
            return False
    if not settings.oxts and "oxts" in entry.name:
        return False
    if not settings.infos and "infos" in entry.name:
        return False
    if not settings.vehicle_data and "vehicle_data" in entry.name:
        return False
    if "lidar" in entry.name:
        if not settings.lidar:
            return False
        if "after" in entry.name or "before" in entry.name:
            # this is only the case for frames, where we have surrounding frames
            distance = int(entry.name.split("_")[2][:2])
            tgt = settings.num_scans_before if "before" in entry.name else settings.num_scans_after
            if tgt != -1 and distance > tgt:
                return False
    return True


def _download_dataset(dl_settings: DownloadSettings, filter_settings: FilterSettings, dirname: str):
    dbx = ResumableDropbox(app_key=APP_KEY, oauth2_refresh_token=REFRESH_TOKEN, timeout=TIMEOUT)
    entries = _list_folder(dl_settings.url, dbx, dirname)
    if not entries:
        typer.echo(
            "Warning! No files found. Check the url, but this could be a known dropbox error.\n"
            "We are working on a fix, so please try again layer. Sorry for the inconvenience."
        )
        return
    dl_dir = osp.join(dl_settings.output_dir, "downloads", dirname)
    files_to_download = [
        DownloadExtractInfo(
            url=dl_settings.url,
            file_path=f"/{dirname}/" + entry.name,
            dl_dir=dl_dir,
            extract_dir=dl_settings.output_dir,
            size=entry.size,
            content_hash=entry.content_hash,
            rm=dl_settings.rm,
            dry_run=dl_settings.dry_run,
            extract=dl_settings.extract,
            extract_already_downloaded=dl_settings.extract_already_downloaded,
        )
        for entry in entries
        if _filter_entry(entry, filter_settings)
    ]
    if not files_to_download:
        typer.echo("No files to download. Perhaps you specified too many filters?")
        return

    if not dl_settings.dry_run:
        os.makedirs(dl_settings.output_dir, exist_ok=True)
        os.makedirs(dl_dir, exist_ok=True)
    if dl_settings.parallel:
        with ThreadPoolExecutor(max_workers=dl_settings.max_workers) as pool:
            pool.map(lambda info: _download_and_extract(dbx, info), files_to_download)
    else:
        for info in files_to_download:
            _download_and_extract(dbx, info)


def _list_folder(url, dbx: ResumableDropbox, path: str):
    shared_link = dropbox.files.SharedLink(url=url)
    try:
        res = dbx.files_list_folder(path=f"/{path}", shared_link=shared_link)
    except dropbox.exceptions.ApiError as err:
        raise click.exceptions.ClickException(
            f"Dropbox raised the following error:\n\t{err}\nThis could be due to:"
            '\n\ta) bad url. Please try it in a browser and specify with quotes (--url="<url>").'
            "\n\tb) zod bandwidth limit. Sorry about this, and please try again the next day."
            "\n\tc) other error (bad internet connection, dropbox outage, etc.)."
        )

    entries = res.entries
    while res.has_more:
        res = dbx.files_list_folder_continue(res.cursor)
        entries.extend(res.entries)
    return entries


def _print_summary(download_settings, filter_settings, subset):
    print("\nDownload settings:")
    print(download_settings)
    print("\nFilter settings:")
    if filter_settings.version == Version.MINI:
        print("    version: mini\n    (other settings are ignored for mini)")
    else:
        print(filter_settings)
        if subset == SubDataset.FRAMES and (filter_settings.num_scans_before == filter_settings.num_scans_after == 0):
            typer.secho(
                "Note! The current settings will only download the core lidar frames. "
                "If you need surrounding scans, set --num-scans-before and/or --num-scans-after.",
                fg=typer.colors.YELLOW,
            )
    print("\n")


REQ = "Required Arguments"
GEN = "General Download Settings"
FIL = "General Filter Settings"
FRA = "Frames Filter Settings"
SEQ = "Sequences/Drives Filter Settings"


@app.command()
def download(
    url: str = typer.Option(
        ...,
        help="The dropbox shared folder url",
        prompt="What is the dropbox url to the dataset (you can get it from zod.zenseact.com)?",
        prompt_required=False,
        rich_help_panel=REQ,
    ),
    output_dir: str = typer.Option(
        ...,
        help="Output directory where dataset will be extracted",
        prompt="Where do you want to extract the dataset (e.g. ~/data/zod)?",
        prompt_required=False,
        rich_help_panel=REQ,
    ),
    subset: SubDataset = typer.Option(
        ...,
        help="The sub-dataset to download",
        prompt="Which sub-dataset do you want to download?",
        prompt_required=False,
        rich_help_panel=REQ,
    ),
    version: Version = typer.Option(
        ...,
        help="The version of the dataset to download",
        prompt="Which version do you want to download?",
        prompt_required=False,
        rich_help_panel=REQ,
    ),
    # General download settings
    rm: bool = typer.Option(False, help="Remove the downloaded archives", rich_help_panel=GEN),
    dry_run: bool = typer.Option(False, help="Print what would be downloaded", rich_help_panel=GEN),
    extract: bool = typer.Option(True, help="Unpack the archives", rich_help_panel=GEN),
    extract_already_downloaded: bool = typer.Option(
        False, help="Extract already downloaded archives", rich_help_panel=GEN
    ),
    parallel: bool = typer.Option(True, help="Download files in parallel", rich_help_panel=GEN),
    max_workers: int = typer.Option(None, help="Max number of workers for parallel downloads", rich_help_panel=GEN),
    no_confirm: bool = typer.Option(
        False,
        "-y",
        "--no-confirm/--confirm",
        help="Don't ask for confirmation",
        is_flag=True,
        flag_value=False,
        rich_help_panel=GEN,
    ),
    # Filter settings
    annotations: bool = typer.Option(True, help="Download annotations", rich_help_panel=FIL),
    images: bool = typer.Option(True, help="Whether to download the images", rich_help_panel=FIL),
    blur: bool = typer.Option(True, help="Download blur images", rich_help_panel=FIL),
    lidar: bool = typer.Option(True, help="Download lidar data", rich_help_panel=FIL),
    oxts: bool = typer.Option(True, help="Download oxts data", rich_help_panel=FIL),
    infos: bool = typer.Option(True, help="Download infos", rich_help_panel=FIL),
    vehicle_data: bool = typer.Option(True, help="Download the vehicle data", rich_help_panel=SEQ),
    dnat: bool = typer.Option(False, help="Download DNAT images", rich_help_panel=FRA),
    num_scans_before: int = typer.Option(
        0, help="Number of earlier lidar scans to download (-1 == all)", rich_help_panel=FRA
    ),
    num_scans_after: int = typer.Option(
        0, help="Number of later lidar scans to download (-1 == all)", rich_help_panel=FRA
    ),
):
    """Download the Zenseact Open Dataset."""
    download_settings = DownloadSettings(
        url=url,
        output_dir=os.path.expanduser(output_dir),
        rm=rm,
        dry_run=dry_run,
        extract=extract,
        extract_already_downloaded=extract_already_downloaded,
        parallel=parallel,
        max_workers=max_workers,
    )
    filter_settings = FilterSettings(
        version=version,
        annotations=annotations,
        images=images,
        blur=blur,
        dnat=dnat,
        lidar=lidar,
        oxts=oxts,
        infos=infos,
        vehicle_data=vehicle_data,
        num_scans_before=num_scans_before if subset == SubDataset.FRAMES else -1,
        num_scans_after=num_scans_after if subset == SubDataset.FRAMES else -1,
    )
    if not no_confirm:
        _print_summary(download_settings, filter_settings, subset)
        typer.confirm(
            f"Download with the above settings?",
            abort=True,
        )
    _download_dataset(download_settings, filter_settings, subset.folder)


if __name__ == "__main__":
    app()



"""Annotation parsers."""

import json
from dataclasses import dataclass
from typing import Any, Dict, List

from zod.anno.lane import LaneAnnotation, parse_lane_annotation
from zod.anno.road_condition import RoadConditionAnnotation
from zod.constants import AnnotationProject
from zod.data_classes._serializable import JSONSerializable

from .ego_road import EgoRoadAnnotation
from .object import ObjectAnnotation
from .tsr.traffic_sign import TrafficSignAnnotation


@dataclass
class AnnotationFile(JSONSerializable):
    """Class to store information about an annotation frame."""

    filepath: str
    project: AnnotationProject

    def read(self) -> Any:
        """Read (and parse) the annotation json."""
        return ANNOTATION_PARSERS[self.project](self.filepath)


def _read_annotation_file(annotation_file: str) -> List[Dict[str, Any]]:
    """Read an annotation file."""
    with open(annotation_file, "r") as file:
        return json.load(file)


def parse_object_detection_annotation(annotation_path: str) -> List[ObjectAnnotation]:
    """Parse the objects annotation from the annotation string."""
    annotation = _read_annotation_file(annotation_path)
    return [ObjectAnnotation.from_dict(obj) for obj in annotation]


def parse_traffic_sign_annotation(annotation_path: str) -> List[TrafficSignAnnotation]:
    """Parse the traffic sign annotation from the annotation string."""
    annotation = _read_annotation_file(annotation_path)
    return [TrafficSignAnnotation.from_dict(sign) for sign in annotation]


def parse_lane_markings_annotation(annotation_path: str) -> List[LaneAnnotation]:
    """Parse the lane markings annotation from the annotation string."""
    annotations = _read_annotation_file(annotation_path)
    return [parse_lane_annotation(marker) for marker in annotations]


def parse_ego_road_annotation(annotation_path: str) -> List[EgoRoadAnnotation]:
    """Parse the egoroad annotation from the annotation string."""
    annotations = _read_annotation_file(annotation_path)
    parsed_annos = []
    for anno in annotations:
        if "EgoRoad" in anno["properties"]["class"]:  # discard Barrier and RoadEdge
            parsed_annos.append(EgoRoadAnnotation.from_dict(anno))
    return parsed_annos


def parse_road_condition_annotation(annotation_path) -> RoadConditionAnnotation:
    with open(annotation_path, "r") as file:
        annotation = json.load(file)
    return RoadConditionAnnotation.from_dict(annotation)


ANNOTATION_PARSERS = {
    AnnotationProject.LANE_MARKINGS: parse_lane_markings_annotation,
    AnnotationProject.EGO_ROAD: parse_ego_road_annotation,
    AnnotationProject.OBJECT_DETECTION: parse_object_detection_annotation,
    AnnotationProject.TRAFFIC_SIGNS: parse_traffic_sign_annotation,
    AnnotationProject.ROAD_CONDITION: parse_road_condition_annotation,
}



from __future__ import annotations

from dataclasses import dataclass
from enum import Enum
from typing import Any, Dict, List


class EgoRoadClass(Enum):
    ROAD = "EgoRoad_Road"
    DEBRIS = "EgoRoad_Debris"


@dataclass
class EgoRoadAnnotation:
    """EgoRoad dataclass."""

    uuid: str
    geometry: List[List[float]]  # [[x1, y1], [x2, y2], ...]
    type: EgoRoadClass

    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> EgoRoadAnnotation:
        return cls(
            uuid=data["properties"]["annotation_uuid"],
            geometry=data["geometry"]["coordinates"],
            type=EgoRoadClass(data["properties"]["class"]),
        )



from ._nuscenes_eval.common.data_classes import EvalBoxes as EvalBoxes
from ._nuscenes_eval.detection.data_classes import DetectionBox as DetectionBox
from .constants import EVALUATION_CLASSES as EVALUATION_CLASSES
from .eval_nuscenes_style import evaluate_nuscenes_style as evaluate_nuscenes_style



# nuScenes dev-kit.
# Code written by Oscar Beijbom and Varun Bankiti, 2019.

from zod.eval.detection.constants import EVALUATION_CLASSES


DETECTION_NAMES = EVALUATION_CLASSES

TP_METRICS = ["trans_err", "scale_err", "orient_err"]

PRETTY_TP_METRICS = {
    "trans_err": "Trans.",
    "scale_err": "Scale",
    "orient_err": "Orient.",
}

TP_METRICS_UNITS = {
    "trans_err": "m",
    "scale_err": "1-IOU",
    "orient_err": "rad.",
}



from datetime import datetime
from typing import Any, List, Optional

from zod.constants import AnnotationProject, Lidar, NotAnnotatedError
from zod.utils.compensation import motion_compensate_scanwise

from .calibration import Calibration
from .ego_motion import EgoMotion
from .info import Information
from .metadata import SequenceMetadata
from .sensor import LidarData
from .vehicle_data import VehicleData


class ZodSequence:
    def __init__(self, info: Information) -> None:
        self.info: Information = info  # holds all the paths to the files
        self._ego_motion: Optional[EgoMotion] = None  # this is the light-weight version of oxts
        self._oxts: Optional[EgoMotion] = None
        self._calibration: Optional[Calibration] = None
        self._metadata: Optional[SequenceMetadata] = None
        self._vehicle_data: Optional[VehicleData] = None

    @property
    def ego_motion(self) -> EgoMotion:
        """Get the oxts file."""
        if self._ego_motion is None:
            self._ego_motion = EgoMotion.from_json_path(self.info.ego_motion_path)
        return self._ego_motion

    @property
    def oxts(self) -> EgoMotion:
        """Get the oxts."""
        if self._oxts is None:
            self._oxts = EgoMotion.from_oxts_path(self.info.oxts_path)
        return self._oxts

    @property
    def calibration(self) -> Calibration:
        """Get the calibration."""
        if self._calibration is None:
            self._calibration = Calibration.from_json_path(self.info.calibration_path)
        return self._calibration

    @property
    def metadata(self) -> SequenceMetadata:
        """Get the metadata."""
        if self._metadata is None:
            self._metadata = SequenceMetadata.from_json_path(self.info.metadata_path)
        return self._metadata

    @property
    def vehicle_data(self) -> VehicleData:
        """Get the vehicle data."""
        if self._vehicle_data is None:
            assert self.info.vehicle_data_path is not None, "Vehicle data path is missing."
            self._vehicle_data = VehicleData.from_hdf5(self.info.vehicle_data_path)
        return self._vehicle_data

    def is_annotated(self, project: AnnotationProject) -> bool:
        """Check if the frame is annotated for a given project."""
        return project in self.info.annotations

    def get_annotation(self, project: AnnotationProject) -> List[Any]:
        """Get the annotation for a given project."""
        if not self.is_annotated(project):
            raise NotAnnotatedError(f"Project {project} is not annotated for sequence {self.info.id}.")
        return self.info.annotations[project].read()

    def get_lidar(self, start: int = 0, end: Optional[int] = None) -> List[LidarData]:
        """Get the point clouds."""
        return [lidar_frame.read() for lidar_frame in self.info.get_lidar_frames(Lidar.VELODYNE)[start:end]]

    def get_compensated_lidar(self, time: datetime) -> LidarData:
        """Get the point cloud at a given timestamp."""
        lid_frame = self.info.get_lidar_frame(time, Lidar.VELODYNE)
        pcd = lid_frame.read()
        return motion_compensate_scanwise(
            pcd,
            self.ego_motion,
            self.calibration.lidars[Lidar.VELODYNE],
            time.timestamp(),
        )

    def get_aggregated_lidar(self, start: int = 0, end: int = None, timestamp: Optional[float] = None) -> LidarData:
        """Get the aggregated point cloud."""
        lidar_scans = self.get_lidar(start, end)
        if timestamp is None:
            timestamp = lidar_scans[len(lidar_scans) // 2].core_timestamp
        for i, scan in enumerate(lidar_scans):
            if scan.core_timestamp == timestamp:
                continue
            compensated_scan = motion_compensate_scanwise(
                scan, self.ego_motion, self.calibration.lidars[Lidar.VELODYNE], timestamp
            )
            lidar_scans[i] = compensated_scan
        aggregated = lidar_scans[0]
        aggregated.extend(*lidar_scans[1:])
        return aggregated

    def get_keyframe_lidar(self, motion_compensated=True) -> LidarData:
        """Get the keyframe point cloud.

        Args:
            motion_compensated (bool, optional): Whether to motion compensate the point cloud
             to camera (and annotation) timestamp. Defaults to True.

        Returns:
            LidarData: The point cloud.
        """

        lidar_scan = self.info.get_key_lidar_frame(Lidar.VELODYNE).read()

        if not motion_compensated:
            return lidar_scan

        return motion_compensate_scanwise(
            lidar_scan,
            self.ego_motion,
            self.calibration.lidars[Lidar.VELODYNE],
            self.info.keyframe_time.timestamp(),
        )



from zod.anno.object import OBJECT_CLASSES, OBJECT_SUBCLASSES

# Try to evaluate all classes, including subclasses.
all_classes = set(OBJECT_CLASSES + OBJECT_SUBCLASSES)
# Remove classes that
all_classes -= {
    "Unclear",
    "Vehicle_Other",
    "VulnerableVehicle_Other",
    "VulnerableVehicle_Wheelchair",
}
EVALUATION_CLASSES = sorted(all_classes)



from __future__ import annotations

from abc import ABC
from dataclasses import dataclass
from typing import Any, Dict, List, Optional

LANE_MARKING_TYPES = []  # TODO


@dataclass
class LaneAnnotation(ABC):
    """Base class for lane annotations."""

    uuid: str
    geometry: List[List[float]]  # [[x1, y1], [x2, y2], ...]


@dataclass
class LaneMarkingAnnotation(LaneAnnotation):
    """Class to store lane marking information."""

    type: str  # e.g. "solid" or "dashed"
    colored: bool
    instance_id: Optional[int]
    cardinality: str

    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> LaneMarkingAnnotation:
        """Create an AnnotatedLaneMarking from a dictionary."""
        properties: Dict[str, Any] = data["properties"]
        if "InstanceID" not in properties or "MultipleLaneMarkings" not in properties:
            pass
        return cls(
            uuid=properties["annotation_uuid"],
            geometry=data["geometry"]["coordinates"],
            type=properties["class"],
            colored=properties["coloured"],
            instance_id=properties.get("InstanceID"),
            cardinality=properties.get("MultipleLaneMarkings", "Single"),
        )


@dataclass
class ShadedAreaAnnotation(LaneAnnotation):
    """Class to store shaded area information."""

    type: str
    colored: bool

    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> ShadedAreaAnnotation:
        """Create an AnnotatedLaneMarking from a dictionary."""
        properties: Dict[str, Any] = data["properties"]
        return cls(
            uuid=properties["annotation_uuid"],
            geometry=data["geometry"]["coordinates"],
            type=properties["ShadedAreaType"],
            colored=properties["coloured"],
        )


@dataclass
class RoadPaintingAnnotation(LaneAnnotation):
    """Class to store road painting information."""

    unclear: Optional[bool]
    odd: Optional[bool]
    contains_arrow: Optional[bool]
    contains_pictogram: Optional[bool]
    contains_text: Optional[bool]
    contains_trafficsign: Optional[bool]
    contains_crosswalk: Optional[bool]
    contains_marker: Optional[bool]
    contains_other: Optional[bool]

    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> RoadPaintingAnnotation:
        """Create an AnnotatedRoadPainting from a dictionary."""
        properties: Dict[str, Any] = data["properties"]

        return cls(
            uuid=properties["annotation_uuid"],
            geometry=data["geometry"]["coordinates"],
            unclear=properties.get("Unclear", None),
            odd=properties.get("Odd", None),
            contains_arrow=properties.get("ContainsArrow", None),
            contains_pictogram=properties.get("ContainsPictogram", None),
            contains_text=properties.get("ContainsText", None),
            contains_trafficsign=properties.get("ContainsTrafficSign", None),
            contains_crosswalk=properties.get("ContainsCrossWalk", None),
            contains_marker=properties.get("ContainsMarker", None),
            contains_other=properties.get("ContainsOther", None),
        )

    @property
    def type(self) -> str:
        """Return the type of road painting."""
        assert (
            sum(
                [
                    bool(self.contains_arrow),
                    bool(self.contains_pictogram),
                    bool(self.contains_text),
                    bool(self.contains_trafficsign),
                    bool(self.contains_crosswalk),
                    bool(self.contains_marker),
                    bool(self.contains_other),
                ]
            )
            <= 1
        ), f"More than one type of road painting is set: {self}"
        if self.contains_arrow:
            return "arrow"
        elif self.contains_pictogram:
            return "pictogram"
        elif self.contains_text:
            return "text"
        elif self.contains_trafficsign:
            return "trafficsign"
        elif self.contains_crosswalk:
            return "crosswalk"
        elif self.contains_marker:
            return "marker"
        elif self.contains_other:
            return "other"
        elif self.unclear:
            print("no road painting type set, but unclear is set", flush=True)
            return "unclear"
        elif self.odd:
            print("no road painting type set, but odd is set", flush=True)
            return "odd"
        else:
            print("ERROR!", flush=True)
            raise ValueError(f"no road painting type set: {self}")


def parse_lane_annotation(data: Dict) -> LaneAnnotation:
    """Parse a lane annotation from a dictionary."""
    if "class" in data["properties"]:
        if data["properties"]["class"] == "shaded_area":
            return ShadedAreaAnnotation.from_dict(data)
        else:
            return LaneMarkingAnnotation.from_dict(data)
    return RoadPaintingAnnotation.from_dict(data)



import os.path as osp
from dataclasses import dataclass
from datetime import datetime
from itertools import chain
from typing import Dict, Iterator, List, Optional, Tuple

from zod.anno.parser import AnnotationFile
from zod.constants import AnnotationProject, Anonymization, Camera, Lidar, Radar

from ._serializable import JSONSerializable
from .sensor import CameraFrame, LidarFrame, RadarFrames, SensorFrame


@dataclass
class Information(JSONSerializable):
    """Base class for frame and sequence information."""

    id: str
    start_time: datetime
    end_time: datetime
    keyframe_time: datetime

    calibration_path: str
    ego_motion_path: str
    metadata_path: str
    oxts_path: str
    vehicle_data_path: Optional[str]

    annotations: Dict[AnnotationProject, AnnotationFile]
    camera_frames: Dict[str, List[CameraFrame]]  # key is a combination of Camera and Anonymization
    lidar_frames: Dict[Lidar, List[LidarFrame]]
    radar_frames: Optional[Dict[Radar, List[RadarFrames]]] = None

    @property
    def all_frames(self) -> Iterator[SensorFrame]:
        """Iterate over all frames."""
        return chain(
            *self.lidar_frames.values(),
            *self.camera_frames.values(),
        )

    def convert_paths_to_absolute(self, root_path: str):
        self.calibration_path = osp.join(root_path, self.calibration_path)
        self.ego_motion_path = osp.join(root_path, self.ego_motion_path)
        self.metadata_path = osp.join(root_path, self.metadata_path)
        self.oxts_path = osp.join(root_path, self.oxts_path)
        if self.vehicle_data_path is not None:
            self.vehicle_data_path = osp.join(root_path, self.vehicle_data_path)
        for annotation in self.annotations.values():
            annotation.filepath = osp.join(root_path, annotation.filepath)
        for frame in self.all_frames:
            frame.filepath = osp.join(root_path, frame.filepath)

    def get_camera_lidar_map(
        self,
        anonymization: Anonymization = Anonymization.BLUR,
        camera: Camera = Camera.FRONT,
        lidar: Lidar = Lidar.VELODYNE,
    ) -> Iterator[Tuple[CameraFrame, SensorFrame]]:
        """Iterate over all camera frames and their corresponding lidar frames.

        Args:
            anonymization: The anonymization to use. e.g., blur
            camera: The camera to use. e.g., camera_front_blur
            lidar: The lidar to use. e.g., lidar_velodyne
        Yields:
            A tuple of the camera frame and the closest lidar frame.
        """
        camera_name = f"{camera.value}_{anonymization.value}"
        assert (
            camera_name in self.camera_frames
        ), f"Camera {camera_name} not found. Available cameras: {self.camera_frames.keys()}"
        assert lidar in self.lidar_frames, f"Lidar {lidar} not found. Available lidars: {self.lidar_frames.keys()}"

        for camera_frame in self.camera_frames[camera_name]:
            # Get the closest lidar frame in time
            lidar_frame = min(
                self.lidar_frames[lidar],
                key=lambda lidar_frame: abs(lidar_frame.time - camera_frame.time),
            )
            yield camera_frame, lidar_frame

    ### Keyframe accessors ###

    def get_key_camera_frame(
        self,
        anonymization: Anonymization = Anonymization.BLUR,
        camera: Camera = Camera.FRONT,
    ) -> CameraFrame:
        camera_name = f"{camera.value}_{anonymization.value}"
        if len(self.camera_frames[camera_name]) == 1:
            return self.camera_frames[camera_name][0]
        else:
            return min(
                self.camera_frames[camera_name],
                key=lambda camera_frame: abs(camera_frame.time - self.keyframe_time),
            )

    def get_key_lidar_frame(self, lidar: Lidar = Lidar.VELODYNE) -> LidarFrame:
        return self.get_lidar_frame(self.keyframe_time, lidar)

    ### Timestamp accessors ###

    def get_camera_frame(
        self,
        time: datetime,
        anonymization: Anonymization = Anonymization.BLUR,
        camera: Camera = Camera.FRONT,
    ) -> CameraFrame:
        camera_name = f"{camera.value}_{anonymization.value}"
        return min(
            self.camera_frames[camera_name],
            key=lambda camera_frame: abs(camera_frame.time - time),
        )

    def get_lidar_frame(self, time: datetime, lidar: Lidar = Lidar.VELODYNE) -> LidarFrame:
        return min(
            self.lidar_frames[lidar],
            key=lambda lidar_frame: abs(lidar_frame.time - time),
        )

    ### Full accessors ###

    def get_camera_frames(
        self,
        anonymization: Anonymization = Anonymization.BLUR,
        camera: Camera = Camera.FRONT,
    ) -> List[CameraFrame]:
        camera_name = f"{camera.value}_{anonymization.value}"
        return self.camera_frames[camera_name]

    def get_lidar_frames(self, lidar: Lidar = Lidar.VELODYNE) -> List[LidarFrame]:
        return self.lidar_frames[lidar]

    def get_radar_frames(self, radar: Radar) -> RadarFrames:
        if self.radar_frames is None:
            err = "No radar frames available!"
            err += "\nPlease download the latest version of ZOD to access radar data. "
            err += "\nRadar is available for ZOD Sequences and Drives only."
            raise ValueError(err)
        return self.radar_frames[radar]



try:
    import typer
except ImportError:
    print('Warning! zod is installed without the CLI dependencies:\npip install "zod[cli]"')
    exit(1)

from zod.cli.download import app as download_app
from zod.cli.extract_tsr_patches import extract_tsr_patches
from zod.cli.generate_coco_json import convert_to_coco
from zod.cli.verify import app as verify_app
from zod.cli.visualize_lidar import app as visualize_lidar_app

visualize_app = typer.Typer(help="Visualize ZOD.", no_args_is_help=True)
visualize_app.add_typer(visualize_lidar_app, name="lidar")

convert_app = typer.Typer(help="Convert ZOD to a different format.", no_args_is_help=True)
convert_app.command("coco", no_args_is_help=True)(convert_to_coco)
convert_app.command("tsr-patches", no_args_is_help=True)(extract_tsr_patches)


def add_child(parent: typer.Typer, child: typer.Typer, name: str):
    """Workaround to handle single-command sub-apps."""
    if len(child.registered_commands) == 1:
        parent.command(name)(child.registered_commands[0].callback)
    else:
        parent.add_typer(child, name=name)


app = typer.Typer(help="Zenseact Open Dataset CLI.", no_args_is_help=True)

add_child(app, download_app, "download")
add_child(app, verify_app, "verify")
add_child(app, visualize_app, "visualize")
add_child(app, convert_app, "generate")


if __name__ == "__main__":
    app()



import os
from collections import defaultdict
from functools import partial
from pathlib import Path
from typing import Dict, List, Union

import typer
from tqdm.contrib.concurrent import process_map

from zod.cli.utils import SubDataset, Version
from zod.data_classes.info import Information

app = typer.Typer(help="ZOD Download Verifyer", no_args_is_help=True)


def _verify_info(info: Information, separate_lidar: bool) -> Dict[str, Dict[str, Union[bool, List[bool]]]]:
    """Verify the given infos."""
    stats = defaultdict(dict)
    stats["general"] = {
        "calibration": os.path.exists(info.calibration_path),
        "ego_motion": os.path.exists(info.ego_motion_path),
        "metadata": os.path.exists(info.metadata_path),
        "oxts": os.path.exists(info.oxts_path),
    }
    if info.vehicle_data_path is not None:
        stats["general"]["vehicle_data"] = os.path.exists(info.vehicle_data_path)
    for annotation in info.annotations.values():
        stats["annotations"][annotation.project.value] = os.path.exists(annotation.filepath)
    for lidar, lidar_frames in info.lidar_frames.items():
        for i, lidar_frame in enumerate(lidar_frames):
            exists = os.path.exists(lidar_frame.filepath)
            name = lidar.value
            if separate_lidar:
                index = i - len(lidar_frames) // 2
                if index < 0:
                    name += f"_{-index:02d}_before"
                elif index > 0:
                    name += f"_{index:02d}_after"
                else:
                    name += "_core"
                stats["lidar"][name] = exists
            else:
                if name not in stats["lidar"]:
                    stats["lidar"][name] = []
                stats["lidar"][name].append(exists)
    for camera, camera_frames in info.camera_frames.items():
        stats["camera"][camera] = []
        for camera_frame in camera_frames:
            stats["camera"][camera].append(os.path.exists(camera_frame.filepath))
    return stats


def _print_results(verified_infos):
    percentages = []
    groups = sorted(set().union(*verified_infos))
    for group in groups:
        keys = sorted(set().union(*(v[group] for v in verified_infos)))
        stats = {k: [d[group][k] for d in verified_infos if group in d and k in d[group]] for k in keys}
        print(f"\n\n{group.upper():^45}\n{'-' * 55}")
        print(f"{'Data':<20} {'Downloaded (%)':<15} {'Expected (count)':<20}")
        print("-" * 55)
        for data_name, data_stats in stats.items():
            if isinstance(data_stats[0], bool):
                successes = sum(data_stats)
                totals = len(data_stats)
            else:  # list
                successes = sum(sum(substats) for substats in data_stats)
                totals = sum(len(substats) for substats in data_stats)
            percentage = 100 * successes / totals
            print(f"{data_name:<20} {percentage:<15.2f} {totals:<20}")
            percentages.append(percentage)
    if all(p == 100 for p in percentages):
        print("\n\033[92m" + f"All files downloaded succesfully!" + "\033[0m")
    elif all(p == 0 for p in percentages):
        print("\n\033[91m" + f"No files were found" + "\033[0m")
    else:
        # print yellow with average percentage
        avg_percentage = sum(percentages) / len(percentages)
        print(
            "\n\033[93m"
            + f"Found {avg_percentage:.1f}% of the dataset. "
            + "This is could be intentional, look at the table above for full details."
            + "\033[0m"
        )


@app.command(no_args_is_help=True)
def verify(
    dataset_root: Path = typer.Option(..., help="Dataset directory"),
    subset: SubDataset = typer.Option(..., help="The sub-dataset to verify"),
    version: Version = typer.Option(..., help="The version of the dataset to verify"),
):
    """Verify the downloaded files."""
    zod = subset.dataset_cls(str(dataset_root.expanduser()), version.value)
    infos = zod.get_all_infos()
    verified_infos = process_map(
        partial(_verify_info, separate_lidar=subset == SubDataset.FRAMES),
        infos.values(),
        chunksize=1 if version == "mini" else min(len(infos) // 100 + 1, 100),
        desc="Verifying files...",
    )
    _print_results(verified_infos)


if __name__ == "__main__":
    app()



# nuScenes dev-kit.
# Code written by Holger Caesar, 2018.

from typing import Any

import numpy as np
from pyquaternion import Quaternion

from ..common.data_classes import EvalBox

DetectionBox = Any  # Workaround as direct imports lead to cyclic dependencies.


def center_distance(gt_box: EvalBox, pred_box: EvalBox) -> float:
    """
    L2 distance between the box centers (xy only).
    :param gt_box: GT annotation sample.
    :param pred_box: Predicted sample.
    :return: L2 distance.
    """
    return np.linalg.norm(np.array(pred_box.translation[:2]) - np.array(gt_box.translation[:2]))


def yaw_diff(gt_box: EvalBox, eval_box: EvalBox, period: float = 2 * np.pi) -> float:
    """
    Returns the yaw angle difference between the orientation of two boxes.
    :param gt_box: Ground truth box.
    :param eval_box: Predicted box.
    :param period: Periodicity in radians for assessing angle difference.
    :return: Yaw angle difference in radians in [0, pi].
    """
    yaw_gt = quaternion_yaw(Quaternion(gt_box.rotation))
    yaw_est = quaternion_yaw(Quaternion(eval_box.rotation))

    return abs(angle_diff(yaw_gt, yaw_est, period))


def angle_diff(x: float, y: float, period: float) -> float:
    """
    Get the smallest angle difference between 2 angles: the angle from y to x.
    :param x: To angle.
    :param y: From angle.
    :param period: Periodicity in radians for assessing angle difference.
    :return: <float>. Signed smallest between-angle difference in range (-pi, pi).
    """

    # calculate angle difference, modulo to [0, 2*pi]
    diff = (x - y + period / 2) % period - period / 2
    if diff > np.pi:
        diff = diff - (2 * np.pi)  # shift (pi, 2*pi] to (-pi, 0]

    return diff


def scale_iou(sample_annotation: EvalBox, sample_result: EvalBox) -> float:
    """
    This method compares predictions to the ground truth in terms of scale.
    It is equivalent to intersection over union (IOU) between the two boxes in 3D,
    if we assume that the boxes are aligned, i.e. translation and rotation are considered identical.
    :param sample_annotation: GT annotation sample.
    :param sample_result: Predicted sample.
    :return: Scale IOU.
    """
    # Validate inputs.
    sa_size = np.array(sample_annotation.size)
    sr_size = np.array(sample_result.size)
    assert all(sa_size > 0), "Error: sample_annotation sizes must be >0."
    assert all(sr_size > 0), "Error: sample_result sizes must be >0."

    # Compute IOU.
    min_wlh = np.minimum(sa_size, sr_size)
    volume_annotation = np.prod(sa_size)
    volume_result = np.prod(sr_size)
    intersection = np.prod(min_wlh)  # type: float
    union = volume_annotation + volume_result - intersection  # type: float
    iou = intersection / union

    return iou


def quaternion_yaw(q: Quaternion) -> float:
    """
    Calculate the yaw angle from a quaternion.
    Note that this only works for a quaternion that represents a box in lidar or global coordinate frame.
    It does not work for a box in the camera frame.
    :param q: Quaternion of interest.
    :return: Yaw angle in radians.
    """

    # Project into xy plane.
    v = np.dot(q.rotation_matrix, np.array([1, 0, 0]))

    # Measure yaw using arctan.
    yaw = np.arctan2(v[1], v[0])

    return yaw


def cummean(x: np.array) -> np.array:
    """
    Computes the cumulative mean up to each position in a NaN sensitive way
    - If all values are NaN return an array of ones.
    - If some values are NaN, accumulate arrays discording those entries.
    """
    if sum(np.isnan(x)) == len(x):
        # Is all numbers in array are NaN's.
        return np.ones(len(x))  # If all errors are NaN set to error to 1 for all operating points.
    else:
        # Accumulate in a nan-aware manner.
        sum_vals = np.nancumsum(x.astype(float))  # Cumulative sum ignoring nans.
        count_vals = np.cumsum(~np.isnan(x))  # Number of non-nans up to each position.
        return np.divide(sum_vals, count_vals, out=np.zeros_like(sum_vals), where=count_vals != 0)



from typing import Dict, Union

from zod._zod_dataset import ZodDataset
from zod.constants import SEQUENCES, TRAINVAL_FILES
from zod.data_classes.sequence import ZodSequence


class ZodSequences(ZodDataset):
    """ZOD Sequences."""

    def __getitem__(self, frame_id: Union[int, str, slice]) -> ZodSequence:
        """Get frame by id, which is a 6-digit zero-padded number. Ex: '000001'."""
        info = super().__getitem__(frame_id)
        return ZodSequence(info)

    @property
    def trainval_files(self) -> Dict[str, str]:
        return TRAINVAL_FILES[SEQUENCES]



from enum import Enum
from typing import Type

from zod import ZodDataset, ZodDrives, ZodFrames, ZodSequences


class SubDataset(Enum):
    FRAMES = "frames"
    SEQUENCES = "sequences"
    DRIVES = "drives"

    @property
    def folder(self) -> str:
        if self == SubDataset.FRAMES:
            return "single_frames"
        else:
            return self.value

    @property
    def dataset_cls(self) -> Type[ZodDataset]:
        return _cls_map[self]


_cls_map = {
    SubDataset.FRAMES: ZodFrames,
    SubDataset.SEQUENCES: ZodSequences,
    SubDataset.DRIVES: ZodDrives,
}


class Version(Enum):
    FULL = "full"
    MINI = "mini"



"""Relevant constants for ZOD."""

import typing
from enum import Enum
from typing import Literal, Union

# Dataset paths
FRAMES = "single_frames"
SEQUENCES = "sequences"
DRIVES = "drives"
DB_DATE_STRING_FORMAT_W_MICROSECONDS = "%Y-%m-%dT%H:%M:%S.%f%z"

Split = Literal["train", "val", "blacklisted"]
SPLITS = typing.get_args(Split)
TRAIN, VAL, BLACKLISTED = SPLITS

Version = Literal["full", "mini"]
VERSIONS = typing.get_args(Version)
FULL, MINI = VERSIONS

TRAINVAL_FILES = {
    FRAMES: {
        FULL: f"trainval-frames-full.json",
        MINI: f"trainval-frames-mini.json",
    },
    SEQUENCES: {
        FULL: f"trainval-sequences-full.json",
        MINI: f"trainval-sequences-mini.json",
    },
    DRIVES: {
        FULL: f"trainval-drives-full.json",
        MINI: f"trainval-drives-mini.json",
    },
}
SPLIT_FILES = {
    FRAMES: {
        FULL: {
            TRAIN: "splits/frames_full_train.txt",
            VAL: "splits/frames_full_val.txt",
        },
        MINI: {
            TRAIN: "splits/frames_mini_train.txt",
            VAL: "splits/frames_mini_val.txt",
        },
    },
    SEQUENCES: {
        FULL: {
            TRAIN: "splits/sequences_full_train.txt",
            VAL: "splits/sequences_full_val.txt",
        },
        MINI: {
            TRAIN: "splits/sequences_mini_train.txt",
            VAL: "splits/sequences_mini_val.txt",
        },
    },
    DRIVES: {
        FULL: {
            TRAIN: "splits/drives_full_train.txt",
            VAL: "splits/drives_full_val.txt",
        },
        MINI: {
            TRAIN: "splits/drives_mini_train.txt",
            VAL: "splits/drives_mini_val.txt",
        },
    },
}


class Anonymization(Enum):
    BLUR = "blur"
    DNAT = "dnat"
    ORIGINAL = "original"


class AnnotationProject(Enum):
    OBJECT_DETECTION = "object_detection"
    LANE_MARKINGS = "lane_markings"
    TRAFFIC_SIGNS = "traffic_signs"
    EGO_ROAD = "ego_road"
    ROAD_CONDITION = "road_condition"


class NotAnnotatedError(Exception):
    """Error that is raised when a project is not annotated for the requested id."""


### Coordinate Frames ###


class Camera(Enum):
    FRONT = "front"


class Lidar(Enum):
    VELODYNE = "velodyne"


class Radar(Enum):
    FRONT = "front"


Ego = Literal["ego"]
EGO = typing.get_args(Ego)[0]
CoordinateFrame = Union[Camera, Lidar, Radar, Ego]


# ZodFrame properties
FRAMES_IMAGE_MEAN = [0.337, 0.345, 0.367]
FRAMES_IMAGE_STD = [0.160, 0.180, 0.214]



from __future__ import annotations

from dataclasses import dataclass


@dataclass
class RoadConditionAnnotation:
    wet: bool
    snowy: bool

    @classmethod
    def from_dict(cls, data: dict) -> RoadConditionAnnotation:
        return RoadConditionAnnotation(
            wet=data["wetness"],
            snowy=data["snow_coverage"],
        )



from __future__ import annotations

from dataclasses import dataclass
from typing import Any, Dict, Optional

from zod.anno.tsr.class_map import get_class_idx
from zod.constants import Camera
from zod.data_classes.box import Box2D


@dataclass
class TrafficSignAnnotation:
    """Traffic sign dataclass."""

    bounding_box: Box2D
    unclear: bool
    uuid: str

    # These are not set if the object is unclear
    traffic_sign_class: Optional[str] = None
    traffic_sign_idx: Optional[int] = None
    occlusion_ratio: Optional[str] = None
    annotation_uuid: Optional[str] = None
    electronic_sign: Optional[bool] = None

    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> TrafficSignAnnotation:
        """Create a TrafficSignAnnotation from a dict."""
        bbox = Box2D.from_points(data["geometry"]["coordinates"], Camera.FRONT)
        if data["properties"]["class"] == "unclear":
            return cls(
                bounding_box=bbox,
                unclear=True,
                uuid=data["properties"]["annotation_uuid"],
            )
        else:
            return cls(
                unclear=data["properties"]["unclear"],
                bounding_box=Box2D.from_points(data["geometry"]["coordinates"], Camera.FRONT),
                traffic_sign_class=data["properties"]["class"],
                traffic_sign_idx=get_class_idx(data["properties"]["class"]),
                occlusion_ratio=data["properties"]["occlusion_ratio"],
                annotation_uuid=data["properties"]["annotation_uuid"],
                electronic_sign=data["properties"]["is_electronic"],
                uuid=data["properties"]["annotation_uuid"],
            )



"""Backwards compatibility for oxts.py."""

import warnings

from .ego_motion import *

warnings.warn(
    "The 'zod.data_classes.oxts' module has been deprecated and will be removed in a future "
    "version. Please use the 'zod.data_classes.ego_motion' module instead.",
    DeprecationWarning,
    stacklevel=2,
)



"""ZOD Frames module."""

from collections import defaultdict
from typing import Dict, Union

from tqdm import tqdm

from zod._zod_dataset import ZodDataset
from zod.constants import FRAMES, TRAINVAL_FILES, AnnotationProject
from zod.data_classes.frame import ZodFrame


class ZodFrames(ZodDataset):
    """ZOD Frames."""

    def __getitem__(self, frame_id: Union[int, str, slice]) -> ZodFrame:
        """Get frame by id, which is a 6-digit zero-padded number. Ex: '000001'."""
        info = super().__getitem__(frame_id)
        return ZodFrame(info)

    @property
    def trainval_files(self) -> Dict[str, str]:
        return TRAINVAL_FILES[FRAMES]

    def get_subclass_counts(self, require_3d=False) -> Dict[str, int]:
        """Will scan the dataset and return all subclasses and their counts."""
        subclasses = defaultdict(int)
        for frame in tqdm(self, desc="Processing annotations..."):
            for obj in frame.get_annotation(AnnotationProject.OBJECT_DETECTION):
                if require_3d and obj.box3d is None:
                    continue
                subclasses[obj.subclass] += 1
        return subclasses



"""Top-level package for Zenseact Open Dataset (ZOD)."""

import importlib.metadata as importlib_metadata

from ._zod_dataset import ZodDataset as ZodDataset  # For type hinting
from .anno.ego_road import EgoRoadAnnotation as EgoRoadAnnotation
from .anno.lane import LaneAnnotation as LaneAnnotation
from .anno.object import ObjectAnnotation as ObjectAnnotation
from .anno.road_condition import RoadConditionAnnotation as RoadConditionAnnotation
from .anno.tsr.traffic_sign import TrafficSignAnnotation as TrafficSignAnnotation
from .constants import AnnotationProject as AnnotationProject
from .constants import Anonymization as Anonymization
from .constants import Camera as Camera
from .constants import Lidar as Lidar
from .data_classes.frame import ZodFrame as ZodFrame
from .data_classes.sequence import ZodSequence as ZodSequence
from .zod_drives import ZodDrives as ZodDrives
from .zod_frames import ZodFrames as ZodFrames
from .zod_sequences import ZodSequences as ZodSequences

try:
    __version__ = importlib_metadata.version(__package__ or __name__)
except importlib_metadata.PackageNotFoundError:
    __version__ = "0.0.0"



"""Script that prepares the dataset for the traffic sign classification task.

Will create a dataset from the original zod frames by cropping out the relevant traffic signs
in each image and storing them for later use.
"""

import json
import os
from dataclasses import dataclass
from itertools import repeat
from pathlib import Path
from typing import Any, Dict, List, Optional, Set, Tuple

import numpy as np
import typer
from PIL import Image
from tqdm.contrib.concurrent import process_map

import zod.anno.parser as parser
import zod.constants as constants
from zod import ZodFrames
from zod.cli.utils import Version
from zod.data_classes.frame import ZodFrame

SUMMARY = """
Finished creating dataset.
.....................................
Created {} traffic sign samples.
Saved dataset to {}, with a dataset.json file that contains the metadata.

Each image is padded with a factor of {}. Note that the padding might be limited
    by the original image size. The padding is stored in the dataset.json file.

Images are saved in the following folder structure:
    <output_folder>/<traffic_sign_class>/<frame_id>_<annotation_uuid>.png
"""


@dataclass
class Settings:
    output_dir: Path
    padding_factor: Optional[float]
    padding_px: Optional[Tuple[int, int]]
    overwrite: bool
    exclude_unclear: bool


def extract_tsr_patches(
    dataset_root: Path = typer.Option(
        ...,
        exists=True,
        dir_okay=True,
        writable=False,
        readable=True,
        resolve_path=True,
        help="Path to the root of the ZOD dataset.",
    ),
    output_dir: Path = typer.Option(
        ...,
        exists=False,
        file_okay=False,
        dir_okay=True,
        writable=True,
        readable=True,
        resolve_path=True,
        help="Path to the output directory.",
    ),
    version: str = typer.Option("full", help="Version of the dataset to use. One of: full, small."),
    padding_factor: Optional[float] = typer.Option(None, help="Factor to multiply the padding with."),
    padding_px_y: Optional[int] = typer.Option(None, help="Padding in y direction."),
    padding_px_x: Optional[int] = typer.Option(None, help="Padding in x direction."),
    num_workers: Optional[int] = typer.Option(None, help="Number of workers to use."),
    overwrite: bool = typer.Option(False, help="Whether to overwrite existing files."),
    exclude_unclear: bool = typer.Option(False, help="Whether to exclude unclear traffic signs."),
):
    """Create a dataset for traffic sign classification."""
    print("Will create dataset from", dataset_root)
    print("Will save dataset to", output_dir)
    output_dir.mkdir(parents=True, exist_ok=True)

    assert not (
        padding_factor is not None and (padding_px_x is not None or padding_px_y is not None)
    ), "Cannot specify both padding and padding_factor"

    padding = (padding_px_x, padding_px_y) if (padding_px_x is not None and padding_px_y is not None) else None

    settings = Settings(
        output_dir=output_dir,
        padding_factor=padding_factor,
        padding_px=padding,
        overwrite=overwrite,
        exclude_unclear=exclude_unclear,
    )

    zod_frames = ZodFrames(str(dataset_root), version=version.value)
    print(f"Will process {len(zod_frames)} full frames.")

    results = process_map(
        _process_frame,
        zod_frames,
        repeat(settings),
        repeat(zod_frames.get_split(constants.TRAIN)),
        desc="Processing frames in ZOD",
        max_workers=num_workers,
        chunksize=1 if num_workers == 1 else 100,
    )
    # flatten the returned list
    traffic_sign_frames: List[Dict[str, Any]] = [frame for frames in results for frame in frames]

    train_frame_ids = set(zod_frames.get_split(constants.TRAIN))
    val_frame_ids = set(zod_frames.get_split(constants.VAL))

    train_frames = [f for f in traffic_sign_frames if f["frame_id"].split("_")[0] in train_frame_ids]
    val_frames = [f for f in traffic_sign_frames if f["frame_id"].split("_")[0] in val_frame_ids]

    # write it to a json file
    with open(os.path.join(output_dir, "dataset.json"), "w") as f:
        json.dump({constants.TRAIN: train_frames, constants.VAL: val_frames}, f)

    print(
        SUMMARY.format(
            len(traffic_sign_frames),
            output_dir,
            padding_factor if padding_factor is not None else padding,
        )
    )


def _process_frame(frame: ZodFrame, settings: Settings, train_ids: Set[str]) -> List[Dict[str, Any]]:
    """Process a single frame."""

    # not all frames have traffic signs
    if constants.AnnotationProject.TRAFFIC_SIGNS not in frame.info.annotations:
        return []

    traffic_signs: List[parser.TrafficSignAnnotation] = frame.get_annotation(constants.AnnotationProject.TRAFFIC_SIGNS)

    if len(traffic_signs) == 0:
        return []

    new_cropped_frames = []
    # load the image
    image_path = frame.info.get_key_camera_frame(constants.Anonymization.BLUR).filepath
    image = np.array(Image.open(image_path))
    for traffic_sign in traffic_signs:
        if settings.exclude_unclear and traffic_sign.unclear:
            continue
        cls_name = "unclear" if traffic_sign.unclear else traffic_sign.traffic_sign_class
        train_or_val = constants.TRAIN if frame.info.id in train_ids else constants.VAL
        cls_folder = settings.output_dir / train_or_val / cls_name
        cls_folder.mkdir(parents=True, exist_ok=True)

        new_frame_id = f"{frame.info.id}_{traffic_sign.uuid}"
        output_file = os.path.join(
            cls_folder,
            f"{new_frame_id}.png",
        )

        # crop out the correct patch
        cropped_image, padding = traffic_sign.bounding_box.crop_from_image(
            image,
            padding_factor=settings.padding_factor,
            padding=settings.padding_px,
        )

        if settings.overwrite or not os.path.exists(output_file):
            pil_image = Image.fromarray(cropped_image)
            pil_image.save(output_file)
        original_widht = traffic_sign.bounding_box.dimension[0]
        original_height = traffic_sign.bounding_box.dimension[1]
        center_x = padding[0] + original_widht // 2
        center_y = padding[1] + original_height // 2

        # create a frame info file
        new_cropped_frames.append(
            {
                "frame_id": new_frame_id,
                "image_path": os.path.join(cls_folder, f"{new_frame_id}.png"),
                "padding_left": float(padding[0]),
                "padding_top": float(padding[1]),
                "padding_right": float(padding[2]),
                "padding_bottom": float(padding[3]),
                "padded_width": cropped_image.shape[1],
                "padded_height": cropped_image.shape[0],
                "center_x": float(center_x),
                "center_y": float(center_y),
                "original_width": float(traffic_sign.bounding_box.dimension[0]),
                "original_height": float(traffic_sign.bounding_box.dimension[1]),
                "annotation": {key: val for key, val in traffic_sign.__dict__.items() if key != "bounding_box"},
            }
        )

    return new_cropped_frames


if __name__ == "__main__":
    typer.run(extract_tsr_patches)
